# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging
import operator
import os
import sys
import tempfile
import time
from collections import defaultdict
from functools import reduce, wraps
from typing import Callable, Coroutine, Dict, Iterator, List, Optional, Set, Type, Union

from .... import oscar as mo
from ....config import Config
from ....core import ChunkGraph, TileableGraph
from ....core.operand import (
    Fetch,
    FetchShuffle,
    MapReduceOperand,
    ShuffleProxy,
    OperandStage,
)
from ....metrics import Metrics
from ....optimization.logical import OptimizationRecords
from ....oscar.profiling import (
    ProfilingData,
    MARS_ENABLE_PROFILING,
)
from ....typing import TileableType, BandType
from ....utils import build_fetch, Timer
from ...cluster.api import ClusterAPI
from ...lifecycle.api import LifecycleAPI
from ...meta.api import MetaAPI
from ...scheduling import SchedulingAPI
from ...subtask import Subtask, SubtaskResult, SubtaskStatus, SubtaskGraph
from ..core import Task, TaskResult, TaskStatus, new_task_id
from .preprocessor import TaskPreprocessor
from .stage import TaskStageProcessor

logger = logging.getLogger(__name__)

MARS_ENABLE_DUMPING_SUBTASK_GRAPH = int(os.environ.get("MARS_DUMP_SUBTASK_GRAPH", 0))


def _record_error(func: Union[Callable, Coroutine] = None, log_when_error=True):
    assert asyncio.iscoroutinefunction(func)

    @wraps(func)
    async def inner(processor: "TaskProcessor", *args, **kwargs):
        try:
            return await func(processor, *args, **kwargs)
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
            if log_when_error:
                logger.exception("Unexpected error happens in %s", func)
            processor._err_infos.append(sys.exc_info())
            raise

    return inner


class TaskProcessor:
    stage_processors: List[TaskStageProcessor]
    cur_stage_processor: Optional[TaskStageProcessor]

    def __init__(
        self,
        task: Task,
        preprocessor: TaskPreprocessor,
        # APIs
        cluster_api: ClusterAPI,
        lifecycle_api: LifecycleAPI,
        scheduling_api: SchedulingAPI,
        meta_api: MetaAPI,
    ):
        self._task = task
        self._preprocessor = preprocessor

        # APIs
        self._cluster_api = cluster_api
        self._lifecycle_api = lifecycle_api
        self._scheduling_api = scheduling_api
        self._meta_api = meta_api

        if MARS_ENABLE_PROFILING:
            ProfilingData.init(task.task_id)
        elif task.extra_config and task.extra_config.get("enable_profiling"):
            ProfilingData.init(task.task_id, task.extra_config["enable_profiling"])

        self._dump_subtask_graph = False
        if MARS_ENABLE_DUMPING_SUBTASK_GRAPH or (
            task.extra_config and task.extra_config.get("dump_subtask_graph")
        ):
            self._dump_subtask_graph = True

        self.result = TaskResult(
            task_id=task.task_id,
            session_id=task.session_id,
            start_time=time.time(),
            status=TaskStatus.pending,
        )
        self.done = asyncio.Event()
        self.stage_processors = []
        self.cur_stage_processor = None

        self._err_infos = []
        self._chunk_graph_iter = None
        self._raw_tile_context = preprocessor.tile_context.copy()
        self._lifecycle_processed_tileables = set()

        # add metrics
        self._chunk_graph_gen_time = Metrics.gauge(
            "mars.chunk_graph_gen_time_secs",
            "Time consuming in seconds to generate a chunk graph",
            ("session_id", "task_id"),
        )
        self._subtask_graph_gen_time = Metrics.gauge(
            "mars.subtask_graph_gen_time_secs",
            "Time consuming in seconds to generate a subtask graph",
            ("session_id", "task_id", "stage_id"),
        )
        self._task_execution_time = Metrics.gauge(
            "mars.task_execution_time_secs",
            "Time consuming in seconds to execute a task",
            ("session_id", "task_id"),
        )

    @property
    def task_id(self):
        return self._task.task_id

    @property
    def preprocessor(self):
        return self._preprocessor

    @property
    def tileable_graph(self):
        return self._preprocessor.tileable_graph

    @property
    def tile_context(self):
        return self._preprocessor.tile_context

    def get_tiled(self, tileable: TileableType):
        return self._preprocessor.get_tiled(tileable)

    @_record_error
    async def optimize(self) -> TileableGraph:
        # optimization, run it in executor,
        # since optimization may be a CPU intensive operation
        return await asyncio.to_thread(self._preprocessor.optimize)

    @_record_error
    async def incref_fetch_tileables(self):
        # incref fetch tileables in tileable graph to prevent them from deleting
        to_incref_tileable_keys = [
            tileable.op.source_key
            for tileable in self.tileable_graph
            if isinstance(tileable.op, Fetch) and tileable in self._raw_tile_context
        ]
        await self._lifecycle_api.incref_tileables(to_incref_tileable_keys)

    @_record_error
    async def decref_fetch_tileables(self):
        fetch_tileable_keys = [
            tileable.op.source_key
            for tileable in self.tileable_graph
            if isinstance(tileable.op, Fetch) and tileable in self._raw_tile_context
        ]
        await self._lifecycle_api.decref_tileables(fetch_tileable_keys)

    @_record_error
    async def incref_result_tileables(self):
        processed = self._lifecycle_processed_tileables
        # track and incref result tileables if tiled
        tracks = [], []
        for result_tileable in self.tileable_graph.result_tileables:
            if result_tileable in processed:  # pragma: no cover
                continue
            try:
                tiled_tileable = self._preprocessor.get_tiled(result_tileable)
                tracks[0].append(result_tileable.key)
                tracks[1].append(
                    self._lifecycle_api.track.delay(
                        result_tileable.key, [c.key for c in tiled_tileable.chunks]
                    )
                )
                processed.add(result_tileable)
            except KeyError:
                # not tiled, skip
                pass
        if tracks:
            await self._lifecycle_api.track.batch(*tracks[1])
            await self._lifecycle_api.incref_tileables(tracks[0])

    @_record_error
    async def decref_result_tileables(self):
        await self._lifecycle_api.decref_tileables(
            [t.key for t in self._lifecycle_processed_tileables]
        )

    @_record_error
    async def incref_stage(self, stage_processor: "TaskStageProcessor"):
        subtask_graph = stage_processor.subtask_graph
        incref_chunk_keys = []
        for subtask in subtask_graph:
            # for subtask has successors, incref number of successors
            n = subtask_graph.count_successors(subtask)
            for c in subtask.chunk_graph.results:
                incref_chunk_keys.extend([c.key] * n)
            # process reducer, since mapper will generate sub keys
            # we incref (main_key, sub_key) for reducer
            for chunk in subtask.chunk_graph:
                if (
                    isinstance(chunk.op, MapReduceOperand)
                    and chunk.op.stage == OperandStage.reduce
                ):
                    # reducer
                    data_keys = chunk.op.get_dependent_data_keys()
                    incref_chunk_keys.extend(data_keys)
                    # main key incref as well, to ensure existence of meta
                    incref_chunk_keys.extend([key[0] for key in data_keys])
        result_chunks = stage_processor.chunk_graph.result_chunks
        incref_chunk_keys.extend([c.key for c in result_chunks])
        logger.debug("Incref chunks %s for stage", incref_chunk_keys)
        await self._lifecycle_api.incref_chunks(incref_chunk_keys)

    @classmethod
    def _get_decref_stage_chunk_keys(
        cls, stage_processor: "TaskStageProcessor"
    ) -> List[str]:
        decref_chunk_keys = []
        error_or_cancelled = stage_processor.error_or_cancelled()
        if stage_processor.subtask_graph:
            subtask_graph = stage_processor.subtask_graph
            if error_or_cancelled:
                # error or cancel, rollback incref for subtask results
                for subtask in subtask_graph:
                    if subtask.subtask_id in stage_processor.decref_subtask:
                        continue
                    stage_processor.decref_subtask.add(subtask.subtask_id)
                    # if subtask not executed, rollback incref of predecessors
                    for inp_subtask in subtask_graph.predecessors(subtask):
                        for result_chunk in inp_subtask.chunk_graph.results:
                            # for reducer chunk, decref mapper chunks
                            if isinstance(result_chunk.op, ShuffleProxy):
                                for chunk in subtask.chunk_graph:
                                    if isinstance(chunk.op, MapReduceOperand):
                                        data_keys = chunk.op.get_dependent_data_keys()
                                        decref_chunk_keys.extend(data_keys)
                                        decref_chunk_keys.extend(
                                            [key[0] for key in data_keys]
                                        )
                        decref_chunk_keys.extend(
                            [c.key for c in inp_subtask.chunk_graph.results]
                        )
            # decref result of chunk graphs
            decref_chunk_keys.extend(
                [c.key for c in stage_processor.chunk_graph.results]
            )
        return decref_chunk_keys

    @mo.extensible
    @_record_error
    async def decref_stage(self, stage_processor: "TaskStageProcessor"):
        decref_chunk_keys = self._get_decref_stage_chunk_keys(stage_processor)
        logger.debug(
            "Decref chunks %s when stage %s finish",
            decref_chunk_keys,
            stage_processor.stage_id,
        )
        await self._lifecycle_api.decref_chunks(decref_chunk_keys)

    @decref_stage.batch
    @_record_error
    async def decref_stage(self, args_list, kwargs_list):
        decref_chunk_keys = []
        for args, kwargs in zip(args_list, kwargs_list):
            decref_chunk_keys.extend(self._get_decref_stage_chunk_keys(*args, **kwargs))
        logger.debug("Decref chunks %s when stage finish", decref_chunk_keys)
        await self._lifecycle_api.decref_chunks(decref_chunk_keys)

    async def _get_next_chunk_graph(
        self, chunk_graph_iter: Iterator[ChunkGraph]
    ) -> Optional[ChunkGraph]:
        def next_chunk_graph():
            try:
                return next(chunk_graph_iter)
            except StopIteration:
                return

        fut = asyncio.to_thread(next_chunk_graph)
        chunk_graph = await fut
        return chunk_graph

    async def _get_available_band_slots(self) -> Dict[BandType, int]:
        async for bands in self._cluster_api.watch_all_bands():
            if bands:
                return bands

    def _init_chunk_graph_iter(self, tileable_graph: TileableGraph):
        if self._chunk_graph_iter is None:
            self._chunk_graph_iter = iter(self._preprocessor.tile(tileable_graph))

    def _get_chunk_optimization_records(self) -> OptimizationRecords:
        if self._preprocessor.chunk_optimization_records_list:
            return self._preprocessor.chunk_optimization_records_list[-1]

    def _get_tileable_id_to_tileable(self) -> Dict[str, TileableType]:
        tileable_id_to_tileable = dict()
        tileable_graph = self._preprocessor.tileable_graph

        for tileable in tileable_graph:
            tileable_id_to_tileable[str(tileable.key)] = tileable

        return tileable_id_to_tileable

    def _get_tileable_to_subtasks(
        self, subtask_graph: SubtaskGraph
    ) -> Dict[TileableType, List[Subtask]]:
        tileable_to_chunks = defaultdict(set)
        chunk_to_subtasks = dict()

        tileable_graph = self._preprocessor.tileable_graph
        tile_ctx = self._preprocessor.tile_context

        for tileable in tileable_graph:
            if tileable not in tile_ctx:
                continue
            for chunk in tile_ctx[tileable].chunks:
                tileable_to_chunks[tileable].add(chunk.key)
                # register chunk mapping for tiled terminals
                chunk_to_subtasks[chunk.key] = set()

        for subtask in subtask_graph:
            for chunk in subtask.chunk_graph:
                # for every non-fuse chunks (including fused),
                # register subtasks if needed
                if (
                    isinstance(chunk.op, (FetchShuffle, Fetch))
                    or chunk.key not in chunk_to_subtasks
                ):
                    continue
                chunk_to_subtasks[chunk.key].add(subtask)

        tileable_to_subtasks = dict()
        # collect subtasks for tileables
        for tileable, chunk_keys in tileable_to_chunks.items():
            tileable_to_subtasks[tileable] = list(
                reduce(
                    operator.or_,
                    [chunk_to_subtasks[chunk_key] for chunk_key in chunk_keys],
                )
            )
        return tileable_to_subtasks

    @_record_error
    async def get_next_stage_processor(self) -> Optional[TaskStageProcessor]:
        tileable_graph = self._preprocessor.tileable_graph
        self._init_chunk_graph_iter(tileable_graph)

        with Timer() as timer:
            chunk_graph = await self._get_next_chunk_graph(self._chunk_graph_iter)
            if chunk_graph is None:
                # tile finished
                self._preprocessor.done = True
                return
        logger.info(
            "Time consuming to gen a chunk graph is %ss with session id %s, task id %s",
            timer.duration,
            self._task.session_id,
            self._task.task_id,
        )
        self._chunk_graph_gen_time.record(
            timer.duration,
            {"session_id": self._task.session_id, "task_id": self._task.task_id},
        )
        stage_id = new_task_id()
        stage_profiling = ProfilingData[self._task.task_id, "general"].nest(
            f"stage_{stage_id}"
        )
        stage_profiling.set(f"tile({len(chunk_graph)})", timer.duration)

        # gen subtask graph
        available_bands = await self._get_available_band_slots()

        with Timer() as timer:
            subtask_graph = await asyncio.to_thread(
                self._preprocessor.analyze,
                chunk_graph,
                available_bands,
                stage_id=stage_id,
            )
        logger.info(
            "Time consuming to gen a subtask graph is %ss with session id %s, task id %s, stage id %s",
            timer.duration,
            self._task.session_id,
            self._task.task_id,
            stage_id,
        )
        self._subtask_graph_gen_time.record(
            timer.duration,
            {
                "session_id": self._task.session_id,
                "task_id": self._task.task_id,
                "stage_id": stage_id,
            },
        )
        stage_profiling.set(f"gen_subtask_graph({len(subtask_graph)})", timer.duration)

        tileable_to_subtasks = await asyncio.to_thread(
            self._get_tileable_to_subtasks, subtask_graph
        )
        tileable_id_to_tileable = await asyncio.to_thread(
            self._get_tileable_id_to_tileable
        )
        stage_processor = TaskStageProcessor(
            stage_id,
            self._task,
            chunk_graph,
            subtask_graph,
            list(available_bands),
            tileable_to_subtasks,
            tileable_id_to_tileable,
            self._get_chunk_optimization_records(),
            self._scheduling_api,
            self._meta_api,
        )
        return stage_processor

    @_record_error
    async def schedule(self, stage_processor: TaskStageProcessor):
        await stage_processor.run()

    def gen_result(self):
        self.result.status = TaskStatus.terminated
        self.result.end_time = time.time()
        for stage_processor in self.stage_processors:
            if stage_processor.result.error is not None:
                err = stage_processor.result.error
                tb = stage_processor.result.traceback
                self._err_infos.append((type(err), err, tb))
        if self._err_infos:
            # grab the last error
            _, err, tb = self._err_infos[-1]
            self.result.error = err
            self.result.traceback = tb
        cost_time_secs = self.result.end_time - self.result.start_time
        logger.info(
            "Time consuming to execute a task is %ss with session id %s, task id %s",
            cost_time_secs,
            self._task.session_id,
            self._task.task_id,
        )
        self._task_execution_time.record(
            cost_time_secs,
            {"session_id": self._task.session_id, "task_id": self._task.task_id},
        )

    def dump_subtask_graph(self):
        from .graph_visualizer import GraphVisualizer

        try:  # pragma: no cover
            import graphviz
        except ImportError:
            graphviz = None

        dot = GraphVisualizer(self).to_dot()
        directory = tempfile.gettempdir()
        file_name = f"mars-{self.task_id}"
        logger.debug(
            "subtask graph is stored in %s", os.path.join(directory, file_name)
        )
        if graphviz is not None:  # pragma: no cover
            g = graphviz.Source(dot)
            g.view(file_name, directory=directory)
        else:
            with open(os.path.join(directory, file_name), "w") as f:
                f.write(dot)

    def finish(self):
        self.done.set()
        if self._dump_subtask_graph:
            self.dump_subtask_graph()
        if MARS_ENABLE_PROFILING or (
            self._task.extra_config and self._task.extra_config.get("enable_profiling")
        ):
            ProfilingData[self._task.task_id, "general"].set(
                "total", time.time() - self.result.start_time
            )
            serialization = ProfilingData[self._task.task_id, "serialization"]
            if not serialization.empty():
                serialization.set(
                    "total",
                    sum(serialization.values()),
                )
            data = ProfilingData.pop(self._task.task_id)
            self.result.profiling = {
                "supervisor": data,
            }

    def is_done(self) -> bool:
        return self.done.is_set()


class TaskProcessorActor(mo.Actor):
    _task_id_to_processor: Dict[str, TaskProcessor]
    _processed_task_ids: Set[str]
    _cur_processor: Optional[TaskProcessor]

    def __init__(self, session_id: str, task_id: str, task_name: str = None):
        self.session_id = session_id
        self.task_id = task_id
        self.task_name = task_name

        self._task_id_to_processor = dict()
        self._cur_processor = None
        self._subtask_decref_events = dict()

        self._cluster_api = None
        self._meta_api = None
        self._lifecycle_api = None
        self._scheduling_api = None

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(self.address)
        self._scheduling_api = await SchedulingAPI.create(self.session_id, self.address)
        self._meta_api = await MetaAPI.create(self.session_id, self.address)
        self._lifecycle_api = await LifecycleAPI.create(self.session_id, self.address)

    @classmethod
    def gen_uid(cls, session_id: str, task_id: str):
        return f"task_processor_{session_id}_{task_id}"

    async def add_task(
        self,
        task: Task,
        tiled_context: Dict[TileableType, TileableType],
        config: Config,
        task_preprocessor_cls: Type[TaskPreprocessor],
    ):
        task_preprocessor = task_preprocessor_cls(
            task, tiled_context=tiled_context, config=config
        )
        processor = TaskProcessor(
            task,
            task_preprocessor,
            self._cluster_api,
            self._lifecycle_api,
            self._scheduling_api,
            self._meta_api,
        )
        self._task_id_to_processor[task.task_id] = processor

        # tell self to start running
        await self.ref().start.tell()

    def _get_unprocessed_task_processor(self):
        for processor in self._task_id_to_processor.values():
            if processor.result.status == TaskStatus.pending:
                return processor

    async def start(self):
        if self._cur_processor is not None:  # pragma: no cover
            # some processor is running
            return

        processor = self._get_unprocessed_task_processor()
        if processor is None:  # pragma: no cover
            return
        self._cur_processor = processor
        processor.result.status = TaskStatus.running

        incref_fetch, incref_result = False, False
        profiling = ProfilingData[processor.task_id, "general"]
        try:
            # optimization
            with Timer() as timer:
                yield processor.optimize()
            profiling.set("optimize", timer.duration)
            # incref fetch tileables to ensure fetch data not deleted
            with Timer() as timer:
                yield processor.incref_fetch_tileables()
            profiling.set("incref_fetch_tileables", timer.duration)
            incref_fetch = True
            while True:
                with Timer() as stage_timer:
                    stage_processor = yield processor.get_next_stage_processor()
                    if stage_processor is None:
                        break
                    stage_profiling = profiling.nest(
                        f"stage_{stage_processor.stage_id}"
                    )
                    # track and incref result tileables
                    yield processor.incref_result_tileables()
                    incref_result = True
                    # incref stage
                    yield processor.incref_stage(stage_processor)
                    # schedule stage
                    processor.stage_processors.append(stage_processor)
                    processor.cur_stage_processor = stage_processor
                    logger.info("Start new stage with id %s.", stage_processor.stage_id)
                    with Timer() as timer:
                        yield processor.schedule(stage_processor)
                    stage_profiling.set("run", timer.duration)
                stage_profiling.set("total", stage_timer.duration)
                if stage_processor.error_or_cancelled():
                    break
        finally:
            processor.gen_result()
            processor.cur_stage_processor = None
            try:
                # clean ups
                decrefs = []
                error_or_cancelled = False
                for stage_processor in processor.stage_processors:
                    if stage_processor.error_or_cancelled():
                        error_or_cancelled = True
                    decrefs.append(processor.decref_stage.delay(stage_processor))
                yield processor.decref_stage.batch(*decrefs)
                if incref_fetch:
                    # revert fetch incref
                    yield processor.decref_fetch_tileables()
                if incref_result and error_or_cancelled:
                    # revert result incref if error or cancelled
                    yield processor.decref_result_tileables()
            finally:
                processor.finish()
                self._cur_processor = None

    async def wait(self, timeout: int = None):
        fs = [
            processor.done.wait() for processor in self._task_id_to_processor.values()
        ]

        _, pending = yield asyncio.wait(fs, timeout=timeout)
        if not pending:
            raise mo.Return(self.result())
        else:
            [fut.cancel() for fut in pending]

    async def cancel(self):
        if self._cur_processor:
            if not self._cur_processor.cur_stage_processor:
                # still in preprocess
                self._cur_processor.preprocessor.cancel()
            else:
                # otherwise, already in stages, cancel current running stage
                await self._cur_processor.cur_stage_processor.cancel()

    def result(self):
        terminated_result = None
        for processor in self._task_id_to_processor.values():
            if processor.result.status != TaskStatus.terminated:
                return processor.result
            else:
                terminated_result = processor.result
        return terminated_result

    def progress(self):
        tiled_percentage = 0.0
        i = 0
        for processor in self._task_id_to_processor.values():
            # get tileable proportion that is tiled
            tileable_graph = processor.tileable_graph
            tileable_context = processor.tile_context
            tiled_percentage += len(tileable_context) / len(tileable_graph)
            i += 1
        tiled_percentage /= i

        # get progress of stages
        subtask_progress = 0.0
        n_stage = 0
        stage_processors = itertools.chain(
            *(
                processor.stage_processors
                for processor in self._task_id_to_processor.values()
            )
        )
        for stage_processor in stage_processors:
            if stage_processor.subtask_graph is None:  # pragma: no cover
                # generating subtask
                continue
            n_subtask = len(stage_processor.subtask_graph)
            if n_subtask == 0:  # pragma: no cover
                continue
            progress = sum(
                result.progress for result in stage_processor.subtask_results.values()
            )
            progress += sum(
                result.progress
                for subtask_key, result in stage_processor.subtask_snapshots.items()
                if subtask_key not in stage_processor.subtask_results
            )
            subtask_progress += progress / n_subtask
            n_stage += 1
        if n_stage > 0:
            subtask_progress /= n_stage

        return subtask_progress * tiled_percentage

    def get_result_tileables(self):
        processor = list(self._task_id_to_processor.values())[-1]
        tileable_graph = processor.tileable_graph
        result = []
        for result_tileable in tileable_graph.result_tileables:
            tiled = processor.get_tiled(result_tileable)
            result.append(build_fetch(tiled))
        return result

    def get_tileable_graph_as_dict(self):
        processor = list(self._task_id_to_processor.values())[-1]
        tileable_graph = processor.tileable_graph

        node_list = []
        edge_list = []

        visited = set()

        for chunk in tileable_graph:
            if chunk.key in visited:
                continue
            visited.add(chunk.key)

            node_name = str(chunk.op)

            node_list.append({"tileableId": chunk.key, "tileableName": node_name})
            for inp, is_pure_dep in zip(chunk.inputs, chunk.op.pure_depends):
                if inp not in tileable_graph:  # pragma: no cover
                    continue
                edge_list.append(
                    {
                        "fromTileableId": inp.key,
                        "toTileableId": chunk.key,
                        "linkType": 1 if is_pure_dep else 0,
                    }
                )

        graph_dict = {"tileables": node_list, "dependencies": edge_list}
        return graph_dict

    def get_tileable_details(self):
        tileable_to_subtasks = dict()
        subtask_results = dict()

        for processor in self._task_id_to_processor.values():
            for stage in processor.stage_processors:
                tileable_to_subtasks.update(stage.tileable_to_subtasks)

                for subtask, result in stage.subtask_results.items():
                    subtask_results[subtask.subtask_id] = result
                for subtask, result in stage.subtask_snapshots.items():
                    if subtask.subtask_id in subtask_results:
                        continue
                    subtask_results[subtask.subtask_id] = result

        tileable_infos = dict()
        for tileable, subtasks in tileable_to_subtasks.items():
            results = [
                subtask_results.get(
                    subtask.subtask_id,
                    SubtaskResult(
                        progress=0.0,
                        status=SubtaskStatus.pending,
                        stage_id=subtask.stage_id,
                    ),
                )
                for subtask in subtasks
            ]

            # calc progress
            if not results:  # pragma: no cover
                progress = 1.0
            else:
                progress = (
                    1.0 * sum(result.progress for result in results) / len(results)
                )

            # calc status
            statuses = set(result.status for result in results)
            if not results or statuses == {SubtaskStatus.succeeded}:
                status = SubtaskStatus.succeeded
            elif statuses == {SubtaskStatus.cancelled}:
                status = SubtaskStatus.cancelled
            elif statuses == {SubtaskStatus.pending}:
                status = SubtaskStatus.pending
            elif SubtaskStatus.errored in statuses:
                status = SubtaskStatus.errored
            else:
                status = SubtaskStatus.running

            fields = tileable.op._FIELDS
            field_values = tileable.op._FIELD_VALUES
            props = {
                fields[attr_name].tag: value
                for attr_name, value in field_values.items()
                if attr_name not in ("_key", "_id")
                and isinstance(value, (int, float, str))
            }

            tileable_infos[tileable.key] = {
                "progress": progress,
                "subtaskCount": len(results),
                "status": status.value,
                "properties": props,
            }

        return tileable_infos

    def get_tileable_subtasks(self, tileable_id: str, with_input_output: bool):
        returned_subtasks = dict()
        subtask_id_to_types = dict()

        subtask_details = dict()
        subtask_graph = subtask_results = subtask_snapshots = None
        for processor in self._task_id_to_processor.values():
            for stage in processor.stage_processors:
                if tileable_id in stage.tileable_id_to_tileable:
                    tileable = stage.tileable_id_to_tileable[tileable_id]
                    returned_subtasks = {
                        subtask.subtask_id: subtask
                        for subtask in stage.tileable_to_subtasks[tileable]
                    }
                    subtask_graph = stage.subtask_graph
                    subtask_results = stage.subtask_results
                    subtask_snapshots = stage.subtask_snapshots
                    break
            if returned_subtasks:
                break

        if subtask_graph is None:  # pragma: no cover
            return {}

        if with_input_output:
            for subtask in list(returned_subtasks.values()):
                for pred in subtask_graph.iter_predecessors(subtask):
                    if pred.subtask_id in returned_subtasks:  # pragma: no cover
                        continue
                    returned_subtasks[pred.subtask_id] = pred
                    subtask_id_to_types[pred.subtask_id] = "Input"
                for succ in subtask_graph.iter_successors(subtask):
                    if succ.subtask_id in returned_subtasks:  # pragma: no cover
                        continue
                    returned_subtasks[succ.subtask_id] = succ
                    subtask_id_to_types[succ.subtask_id] = "Output"

        for subtask in returned_subtasks.values():
            subtask_result = subtask_results.get(
                subtask,
                subtask_snapshots.get(
                    subtask,
                    SubtaskResult(
                        progress=0.0,
                        status=SubtaskStatus.pending,
                        stage_id=subtask.stage_id,
                    ),
                ),
            )
            subtask_details[subtask.subtask_id] = {
                "name": subtask.subtask_name,
                "status": subtask_result.status.value,
                "progress": subtask_result.progress,
                "nodeType": subtask_id_to_types.get(subtask.subtask_id, "Calculation"),
            }

        for subtask in returned_subtasks.values():
            pred_ids = []
            for pred in subtask_graph.iter_predecessors(subtask):
                if pred.subtask_id in returned_subtasks:
                    pred_ids.append(pred.subtask_id)
            subtask_details[subtask.subtask_id]["fromSubtaskIds"] = pred_ids
        return subtask_details

    def get_result_tileable(self, tileable_key: str):
        processor = list(self._task_id_to_processor.values())[-1]
        tileable_graph = processor.tileable_graph
        for result_tileable in tileable_graph.result_tileables:
            if result_tileable.key == tileable_key:
                tiled = processor.get_tiled(result_tileable)
                return build_fetch(tiled)
        raise KeyError(f"Tileable {tileable_key} does not exist")  # pragma: no cover

    async def _decref_input_subtasks(
        self, subtask: Subtask, subtask_graph: SubtaskGraph
    ):
        # make sure subtasks are decreffed only once
        if subtask.subtask_id not in self._subtask_decref_events:
            self._subtask_decref_events[subtask.subtask_id] = asyncio.Event()
        else:  # pragma: no cover
            await self._subtask_decref_events[subtask.subtask_id].wait()
            return

        decref_chunk_keys = []
        for in_subtask in subtask_graph.iter_predecessors(subtask):
            for result_chunk in in_subtask.chunk_graph.results:
                # for reducer chunk, decref mapper chunks
                if isinstance(result_chunk.op, ShuffleProxy):
                    for chunk in subtask.chunk_graph:
                        if (
                            isinstance(chunk.op, MapReduceOperand)
                            and chunk.op.stage == OperandStage.reduce
                        ):
                            data_keys = chunk.op.get_dependent_data_keys()
                            decref_chunk_keys.extend(data_keys)
                            # decref main key as well
                            decref_chunk_keys.extend([key[0] for key in data_keys])
                decref_chunk_keys.append(result_chunk.key)
        logger.debug(
            "Decref chunks %s when subtask %s finish",
            decref_chunk_keys,
            subtask.subtask_id,
        )
        await self._lifecycle_api.decref_chunks(decref_chunk_keys)

        # `set_subtask_result` will be called when subtask finished
        # but report progress will call set_subtask_result too,
        # so it have risk to duplicate decrease some subtask input object reference,
        # it will cause object reference count lower zero
        # TODO(Catch-Bull): Pop asyncio.Event when current subtask `set_subtask_result`
        # will never be called
        self._subtask_decref_events[subtask.subtask_id].set()

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        logger.debug(
            "Set subtask %s with result %s.", subtask_result.subtask_id, subtask_result
        )
        if (
            self._cur_processor is None
            or self._cur_processor.cur_stage_processor is None
            or (
                subtask_result.stage_id
                and self._cur_processor.cur_stage_processor.stage_id
                != subtask_result.stage_id
            )
        ):
            logger.warning(
                "Stage %s for subtask %s not exists, got stale subtask result %s which may be "
                "speculative execution from previous stages, just ignore it.",
                subtask_result.stage_id,
                subtask_result.subtask_id,
                subtask_result,
            )
            return
        stage_processor = self._cur_processor.cur_stage_processor
        subtask = stage_processor.subtask_id_to_subtask[subtask_result.subtask_id]

        prev_result = stage_processor.subtask_results.get(subtask)
        if prev_result and (
            prev_result.status == SubtaskStatus.succeeded
            or prev_result.progress > subtask_result.progress
        ):
            logger.info(
                "Skip set subtask %s with result %s, previous result is %s.",
                subtask.subtask_id,
                subtask_result,
                prev_result,
            )
            # For duplicate run of subtasks, if the progress is smaller or the subtask has finished or canceled
            # in task speculation, just do nothing.
            # TODO(chaokunyang) If duplicate run of subtasks failed, it may be the fault in worker node,
            #  print the exception, and if multiple failures on the same node, remove the node from the cluster.
            return
        if subtask_result.bands:
            [band] = subtask_result.bands
        else:
            band = None
        stage_processor.subtask_snapshots[subtask] = subtask_result.update(
            stage_processor.subtask_snapshots.get(subtask)
        )
        if subtask_result.status.is_done:
            # update stage_processor.subtask_results to avoid concurrent set_subtask_result
            # since we release lock when `_decref_input_subtasks`.
            stage_processor.subtask_results[subtask] = subtask_result.update(
                stage_processor.subtask_results.get(subtask)
            )
            try:
                # Since every worker will call supervisor to set subtask result,
                # we need to release actor lock to make `decref_chunks` parallel to avoid blocking
                # other `set_subtask_result` calls.
                # If speculative execution enabled, concurrent subtasks may got error since input chunks may
                # got deleted. But it's OK because the current subtask run has succeed.
                if subtask.subtask_id not in stage_processor.decref_subtask:
                    stage_processor.decref_subtask.add(subtask.subtask_id)
                    yield self._decref_input_subtasks(
                        subtask, stage_processor.subtask_graph
                    )

            except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
                logger.debug(
                    "Decref input subtasks for subtask %s failed.", subtask.subtask_id
                )
                _, err, tb = sys.exc_info()
                if subtask_result.status not in (
                    SubtaskStatus.errored,
                    SubtaskStatus.cancelled,
                ):
                    subtask_result.status = SubtaskStatus.errored
                    subtask_result.error = err
                    subtask_result.traceback = tb
            await stage_processor.set_subtask_result(subtask_result, band=band)

    def is_done(self) -> bool:
        for processor in self._task_id_to_processor.values():
            if not processor.is_done():
                return False
        return True



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import sys
import types
from functools import partial, wraps
from typing import Any, Dict, List

import pandas as pd

from ..utils import TypeDispatcher

import cloudpickle

if sys.version_info[:2] < (3, 8):
    try:
        import pickle5 as pickle  # nosec  # pylint: disable=import_pickle
    except ImportError:
        import pickle  # nosec  # pylint: disable=import_pickle
else:
    import pickle  # nosec  # pylint: disable=import_pickle

HAS_PICKLE_BUFFER = pickle.HIGHEST_PROTOCOL >= 5
BUFFER_PICKLE_PROTOCOL = max(pickle.DEFAULT_PROTOCOL, 5)
_PANDAS_HAS_MGR = hasattr(pd.Series([0]), "_mgr")


_serial_dispatcher = TypeDispatcher()
_deserializers = dict()


class Serializer:
    serializer_name = None

    def serialize(self, obj: Any, context: Dict):
        raise NotImplementedError

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        raise NotImplementedError

    @classmethod
    def register(cls, obj_type):
        inst = cls()
        _serial_dispatcher.register(obj_type, inst)
        _deserializers[cls.serializer_name] = inst

    @classmethod
    def unregister(cls, obj_type):
        _serial_dispatcher.unregister(obj_type)
        _deserializers.pop(cls.serializer_name, None)


def buffered(func):
    @wraps(func)
    def wrapped(self, obj: Any, context: Dict):
        if id(obj) in context:
            return {"id": id(obj), "serializer": "ref", "buf_num": 0}, []
        else:
            context[id(obj)] = obj
            return func(self, obj, context)

    return wrapped


def pickle_buffers(obj):
    buffers = [None]
    if HAS_PICKLE_BUFFER:

        def buffer_cb(x):
            x = x.raw()
            if x.ndim > 1:
                # ravel n-d memoryview
                x = x.cast(x.format)
            buffers.append(memoryview(x))

        buffers[0] = cloudpickle.dumps(
            obj,
            buffer_callback=buffer_cb,
            protocol=BUFFER_PICKLE_PROTOCOL,
        )
    else:  # pragma: no cover
        buffers[0] = cloudpickle.dumps(obj)
    return buffers


def unpickle_buffers(buffers):
    result = cloudpickle.loads(buffers[0], buffers=buffers[1:])

    # as pandas prior to 1.1.0 use _data instead of _mgr to hold BlockManager,
    # deserializing from high versions may produce mal-functioned pandas objects,
    # thus the patch is needed
    if _PANDAS_HAS_MGR:
        return result
    else:  # pragma: no cover
        if hasattr(result, "_mgr") and isinstance(result, (pd.DataFrame, pd.Series)):
            result._data = getattr(result, "_mgr")
            delattr(result, "_mgr")
        return result


class ScalarSerializer(Serializer):
    serializer_name = "scalar"

    def serialize(self, obj: Any, context: Dict):
        header = {"val": obj}
        return header, []

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        return header["val"]


class StrSerializer(Serializer):
    serializer_name = "str"

    @buffered
    def serialize(self, obj, context: Dict):
        header = {}
        if isinstance(obj, str):
            header["unicode"] = True
            bytes_data = obj.encode()
        else:
            bytes_data = obj
        return header, [bytes_data]

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        if header.get("unicode"):
            buffer = buffers[0]
            if isinstance(buffer, memoryview):
                buffer = buffer.tobytes()
            return buffer.decode()
        return buffers[0]


class PickleSerializer(Serializer):
    serializer_name = "pickle"

    @buffered
    def serialize(self, obj, context: Dict):
        return {}, pickle_buffers(obj)

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        return unpickle_buffers(buffers)


class CollectionSerializer(Serializer):
    obj_type = None

    @staticmethod
    def _serialize(c):
        headers = [None] * len(c)
        buffers_list = [None] * len(c)
        for idx, obj in enumerate(c):
            header, buffers = yield obj
            headers[idx] = header
            buffers_list[idx] = buffers
        return headers, buffers_list

    @buffered
    def serialize(self, obj: Any, context: Dict):
        buffers = []
        headers_list, buffers_list = yield from self._serialize(obj)
        for b in buffers_list:
            buffers.extend(b)
        headers = {"headers": headers_list}
        if type(obj) is not self.obj_type:
            headers["obj_type"] = pickle.dumps(type(obj))
        return headers, buffers

    @staticmethod
    def _list_deserial(headers: Dict, buffers: List):
        pos = 0
        ret = [None] * len(headers)
        for idx, sub_header in enumerate(headers):
            buf_num = sub_header["buf_num"]
            sub_buffers = buffers[pos : pos + buf_num]
            ret[idx] = yield sub_header, sub_buffers
            pos += buf_num
        return ret

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        obj_type = self.obj_type
        if "obj_type" in header:
            obj_type = pickle.loads(header["obj_type"])
        if hasattr(obj_type, "_fields"):
            # namedtuple
            return obj_type(
                *(yield from self._list_deserial(header["headers"], buffers))
            )
        else:
            return obj_type(
                (yield from self._list_deserial(header["headers"], buffers))
            )


class ListSerializer(CollectionSerializer):
    serializer_name = "list"
    obj_type = list

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        ret = yield from super().deserialize(header, buffers, context)
        for idx, v in enumerate(ret):
            if isinstance(v, Placeholder):
                v.callbacks.append(partial(ret.__setitem__, idx))
        return ret


class TupleSerializer(CollectionSerializer):
    serializer_name = "tuple"
    obj_type = tuple

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        ret = yield from super().deserialize(header, buffers, context)
        assert not any(isinstance(v, Placeholder) for v in ret)
        return ret


class DictSerializer(CollectionSerializer):
    serializer_name = "dict"
    _inspected_inherits = set()

    @buffered
    def serialize(self, obj: Dict, context: Dict):
        obj_type = type(obj)
        if obj_type is not dict and obj_type not in self._inspected_inherits:
            inspect_init = inspect.getfullargspec(obj_type.__init__)
            if (
                inspect_init.args == ["self"]
                and not inspect_init.varargs
                and not inspect_init.varkw
            ):
                # dict inheritance
                # remove context to generate real serialized result
                context.pop(id(obj))
                PickleSerializer.register(obj_type)
                return (yield obj)
            else:
                self._inspected_inherits.add(obj_type)

        key_headers, key_buffers_list = yield from self._serialize(obj.keys())
        value_headers, value_buffers_list = yield from self._serialize(obj.values())

        buffers = []
        for b in key_buffers_list:
            buffers.extend(b)
        key_buf_num = len(buffers)

        for b in value_buffers_list:
            buffers.extend(b)

        header = {
            "key_headers": key_headers,
            "key_buf_num": key_buf_num,
            "value_headers": value_headers,
        }
        if type(obj) is not dict:
            header["obj_type"] = pickle.dumps(type(obj))

        return header, buffers

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        key_buffers = buffers[: header["key_buf_num"]]
        value_buffers = buffers[header["key_buf_num"] :]

        obj_type = dict
        if "obj_type" in header:
            obj_type = pickle.loads(header["obj_type"])

        keys = yield from self._list_deserial(header["key_headers"], key_buffers)
        values = yield from self._list_deserial(header["value_headers"], value_buffers)

        def _key_replacer(key, real_key):
            ret[real_key] = ret.pop(key)

        def _value_replacer(key, real_value):
            if isinstance(key, Placeholder):
                key = context[key.id]
            ret[key] = real_value

        try:
            ret = obj_type(zip(keys, values))
        except TypeError:
            # defaultdict
            ret = obj_type()
            ret.update(zip(keys, values))
        for k, v in zip(keys, values):
            if isinstance(k, Placeholder):
                k.callbacks.append(partial(_key_replacer, k))
            if isinstance(v, Placeholder):
                v.callbacks.append(partial(_value_replacer, k))
        return ret


PickleSerializer.register(object)
ScalarSerializer.register(bool)
ScalarSerializer.register(int)
ScalarSerializer.register(float)
ScalarSerializer.register(complex)
ScalarSerializer.register(type(None))
StrSerializer.register(bytes)
StrSerializer.register(str)
ListSerializer.register(list)
TupleSerializer.register(tuple)
DictSerializer.register(dict)


class Placeholder:
    id: int
    callbacks: List

    def __init__(self, id_: int):
        self.id = id_
        self.callbacks = []

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):  # pragma: no cover
        if not isinstance(other, Placeholder):
            return False
        return self.id == other.id


def serialize(obj, context: Dict = None):
    def _wrap_headers(_obj, _serializer_name, _header, _buffers):
        if _header.get("serializer") == "ref":
            return _header, _buffers
        # if serializer already defined, do not change
        _header["serializer"] = _header.get("serializer", _serializer_name)
        _header["buf_num"] = len(_buffers)
        _header["id"] = id(_obj)
        return _header, _buffers

    context = context if context is not None else dict()

    serializer = _serial_dispatcher.get_handler(type(obj))
    result = serializer.serialize(obj, context)

    if not isinstance(result, types.GeneratorType):
        # result is not a generator, return directly
        header, buffers = result
        return _wrap_headers(obj, serializer.serializer_name, header, buffers)
    else:
        # result is a generator, iter it till final result
        gen_stack = [(result, obj, serializer.serializer_name)]
        last_serial = None
        while gen_stack:
            gen, call_obj, ser_name = gen_stack[-1]
            try:
                gen_to_serial = gen.send(last_serial)
                gen_serializer = _serial_dispatcher.get_handler(type(gen_to_serial))
                gen_result = gen_serializer.serialize(gen_to_serial, context)
                if isinstance(gen_result, types.GeneratorType):
                    # when intermediate result still generator, push its contexts
                    # into stack and handle it first
                    gen_stack.append(
                        (gen_result, gen_to_serial, gen_serializer.serializer_name)
                    )
                    # result need to be emptied to run the generator
                    last_serial = None
                else:
                    # when intermediate result is not generator, pass it
                    # to the generator again
                    last_serial = _wrap_headers(
                        gen_to_serial, gen_serializer.serializer_name, *gen_result
                    )
            except StopIteration as si:
                # when current generator finishes, jump to the previous one
                # and pass final result to it
                gen_stack.pop()
                last_serial = _wrap_headers(call_obj, ser_name, *si.value)

        return last_serial


def deserialize(header: Dict, buffers: List, context: Dict = None):
    def _deserialize(_header, _buffers):
        serializer_name = _header["serializer"]
        obj_id = _header["id"]
        if serializer_name == "ref":
            try:
                result = context[obj_id]
            except KeyError:
                result = context[obj_id] = Placeholder(obj_id)
        else:
            serializer = _deserializers[serializer_name]
            result = serializer.deserialize(_header, _buffers, context)
            if not isinstance(result, types.GeneratorType):
                _fill_context(obj_id, result)
        return result

    def _fill_context(obj_id, result):
        context_val, context[obj_id] = context.get(obj_id), result
        if isinstance(context_val, Placeholder):
            for cb in context_val.callbacks:
                cb(result)

    context = context if context is not None else dict()

    deserialized = _deserialize(header, buffers)
    if not isinstance(deserialized, types.GeneratorType):
        # result is not a generator, return directly
        return deserialized
    else:
        # result is a generator, iter it till final result
        gen_stack = [(deserialized, (header, buffers))]
        last_deserial = None
        while gen_stack:
            gen, to_deserial = gen_stack[-1]
            try:
                gen_to_deserial = gen.send(last_deserial)
                gen_deserialized = _deserialize(*gen_to_deserial)
                if isinstance(gen_deserialized, types.GeneratorType):
                    # when intermediate result still generator, push its contexts
                    # into stack and handle it first
                    gen_stack.append((gen_deserialized, gen_to_deserial))
                    # result need to be emptied to run the generator
                    last_deserial = None
                else:
                    # when intermediate result is not generator, pass it
                    # to the generator again
                    last_deserial = gen_deserialized
            except StopIteration as si:
                # when current generator finishes, jump to the previous one
                # and pass final result to it
                gen_stack.pop()
                last_deserial = si.value
                # remember to fill Placeholders when some result is generated
                _fill_context(to_deserial[0]["id"], last_deserial)
        return last_deserial



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import deque, defaultdict
from typing import Dict, List, Tuple, Type, Union

from ....config import Config
from ....core import ChunkGraph, ChunkType, enter_mode
from ....core.operand import Fetch, VirtualOperand, LogicKeyGenerator
from ....typing import BandType
from ....utils import build_fetch, tokenize
from ...subtask import SubtaskGraph, Subtask
from ..core import Task, new_task_id
from .assigner import AbstractGraphAssigner, GraphAssigner
from .fusion import Coloring

logger = logging.getLogger(__name__)


class GraphAnalyzer:
    def __init__(
        self,
        chunk_graph: ChunkGraph,
        band_slots: Dict[BandType, int],
        task: Task,
        config: Config,
        graph_assigner_cls: Type[AbstractGraphAssigner] = None,
        stage_id: str = None,
    ):
        self._chunk_graph = chunk_graph
        self._band_slots = band_slots
        self._task = task
        self._stage_id = stage_id
        self._config = config
        self._fuse_enabled = task.fuse_enabled
        self._extra_config = task.extra_config
        if graph_assigner_cls is None:
            graph_assigner_cls = GraphAssigner
        self._graph_assigner_cls = graph_assigner_cls
        self._chunk_to_copied = dict()
        self._logic_key_generator = LogicKeyGenerator()

    @classmethod
    def _iter_start_ops(cls, chunk_graph: ChunkGraph):
        visited = set()
        op_keys = set()
        start_chunks = deque(chunk_graph.iter_indep())
        stack = deque([start_chunks.popleft()])

        while stack:
            chunk = stack.popleft()
            if chunk not in visited:
                inp_chunks = chunk_graph.predecessors(chunk)
                if not inp_chunks or all(
                    inp_chunk in visited for inp_chunk in inp_chunks
                ):
                    if len(inp_chunks) == 0:
                        op_key = chunk.op.key
                        if op_key not in op_keys:
                            op_keys.add(op_key)
                            yield chunk.op
                    visited.add(chunk)
                    stack.extend(c for c in chunk_graph[chunk] if c not in visited)
                else:
                    stack.appendleft(chunk)
                    stack.extendleft(
                        reversed(
                            [
                                c
                                for c in chunk_graph.predecessors(chunk)
                                if c not in visited
                            ]
                        )
                    )
            if not stack and start_chunks:
                stack.appendleft(start_chunks.popleft())

    @classmethod
    def _gen_input_chunks(
        cls,
        inp_chunks: List[ChunkType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> List[ChunkType]:
        # gen fetch chunks for input chunks
        inp_fetch_chunks = []
        for inp_chunk in inp_chunks:
            if inp_chunk in chunk_to_fetch_chunk:
                inp_fetch_chunks.append(chunk_to_fetch_chunk[inp_chunk])
            elif isinstance(inp_chunk.op, Fetch):
                chunk_to_fetch_chunk[inp_chunk] = inp_chunk
                inp_fetch_chunks.append(inp_chunk)
            else:
                fetch_chunk = build_fetch(inp_chunk).data
                chunk_to_fetch_chunk[inp_chunk] = fetch_chunk
                inp_fetch_chunks.append(fetch_chunk)

        return inp_fetch_chunks

    @staticmethod
    def _to_band(band_or_worker: Union[BandType, str]) -> BandType:
        if isinstance(band_or_worker, tuple) and len(band_or_worker) == 2:
            # band already
            return band_or_worker
        else:
            return band_or_worker, "numa-0"

    def _gen_subtask_info(
        self,
        chunks: List[ChunkType],
        chunk_to_subtask: Dict[ChunkType, Subtask],
        chunk_to_bands: Dict[ChunkType, BandType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> Tuple[Subtask, List[Subtask]]:
        # gen subtask and its input subtasks
        final_result_chunks_set = set(self._chunk_graph.result_chunks)
        chunks_set = set(chunks)

        result_chunks = []
        result_chunks_set = set()
        chunk_graph = ChunkGraph(result_chunks)
        out_of_scope_chunks = []
        chunk_to_copied = self._chunk_to_copied
        # subtask properties
        band = None
        is_virtual = None
        retryable = True
        chunk_priority = None
        expect_worker = None
        bands_specified = None
        for chunk in chunks:
            if expect_worker is None:
                expect_worker = chunk.op.expect_worker
                bands_specified = expect_worker is not None
            else:  # pragma: no cover
                assert (
                    chunk.op.expect_worker is None
                    or expect_worker == chunk.op.expect_worker
                ), (
                    f"expect_worker {chunk.op.expect_worker} conflicts with chunks that have same color: "
                    f"{expect_worker}"
                )
            # process band
            chunk_band = chunk_to_bands.get(chunk)
            if chunk_band is not None:
                assert (
                    band is None or band == chunk_band
                ), "band conflicts with chunks that have same color"
                band = chunk_band
            # process is_virtual
            if isinstance(chunk.op, VirtualOperand):
                assert is_virtual is None, "only 1 virtual operand can exist"
                is_virtual = True
            else:
                is_virtual = False
            # process retryable
            if not chunk.op.retryable:
                retryable = False
            # process priority
            if chunk.op.priority is not None:
                assert (
                    chunk_priority is None or chunk_priority == chunk.op.priority
                ), "priority conflicts with chunks that have same color"
                chunk_priority = chunk.op.priority
            # process input chunks
            inp_chunks = []
            build_fetch_index_to_chunks = dict()
            for i, inp_chunk in enumerate(chunk.inputs):
                if inp_chunk in chunks_set:
                    inp_chunks.append(chunk_to_copied[inp_chunk])
                else:
                    build_fetch_index_to_chunks[i] = inp_chunk
                    inp_chunks.append(None)
                    if not isinstance(inp_chunk.op, Fetch):
                        out_of_scope_chunks.append(inp_chunk)
            fetch_chunks = self._gen_input_chunks(
                list(build_fetch_index_to_chunks.values()), chunk_to_fetch_chunk
            )
            for i, fetch_chunk in zip(build_fetch_index_to_chunks, fetch_chunks):
                inp_chunks[i] = fetch_chunk
            copied_op = chunk.op.copy()
            copied_op._key = chunk.op.key
            out_chunks = [
                c.data
                for c in copied_op.new_chunks(
                    inp_chunks, kws=[c.params.copy() for c in chunk.op.outputs]
                )
            ]
            for src_chunk, out_chunk in zip(chunk.op.outputs, out_chunks):
                out_chunk._key = src_chunk.key
                chunk_graph.add_node(out_chunk)
                chunk_to_copied[src_chunk] = out_chunk
                if chunk in final_result_chunks_set:
                    result_chunks.append(out_chunk)
                    result_chunks_set.add(out_chunk)
                if not is_virtual:
                    # skip adding fetch chunk to chunk graph when op is virtual operand
                    for c in inp_chunks:
                        if c not in chunk_graph:
                            chunk_graph.add_node(c)
                        chunk_graph.add_edge(c, out_chunk)
        # add chunks with no successors into result chunks
        result_chunks.extend(
            c
            for c in chunk_graph.iter_indep(reverse=True)
            if c not in result_chunks_set
        )
        expect_bands = (
            [self._to_band(expect_worker)]
            if bands_specified
            else ([band] if band is not None else None)
        )
        # calculate priority
        if out_of_scope_chunks:
            inp_subtasks = []
            for out_of_scope_chunk in out_of_scope_chunks:
                copied_out_of_scope_chunk = chunk_to_copied[out_of_scope_chunk]
                inp_subtask = chunk_to_subtask[out_of_scope_chunk]
                if (
                    copied_out_of_scope_chunk
                    not in inp_subtask.chunk_graph.result_chunks
                ):
                    # make sure the chunk that out of scope
                    # is in the input subtask's results,
                    # or the meta may be lost
                    inp_subtask.chunk_graph.result_chunks.append(
                        copied_out_of_scope_chunk
                    )
                inp_subtasks.append(inp_subtask)
            depth = max(st.priority[0] for st in inp_subtasks) + 1
        else:
            inp_subtasks = []
            depth = 0
        priority = (depth, chunk_priority or 0)

        subtask = Subtask(
            subtask_id=new_task_id(),
            stage_id=self._stage_id,
            logic_key=self._gen_logic_key(chunks),
            session_id=self._task.session_id,
            task_id=self._task.task_id,
            chunk_graph=chunk_graph,
            expect_bands=expect_bands,
            bands_specified=bands_specified,
            virtual=is_virtual,
            priority=priority,
            retryable=retryable,
            extra_config=self._extra_config,
        )
        return subtask, inp_subtasks

    def _gen_logic_key(self, chunks: List[ChunkType]):
        return tokenize(
            *[self._logic_key_generator.get_logic_key(chunk.op) for chunk in chunks]
        )

    @enter_mode(build=True)
    def gen_subtask_graph(self) -> SubtaskGraph:
        """
        Analyze chunk graph and generate subtask graph.

        Returns
        -------
        subtask_graph: SubtaskGraph
            Subtask graph.
        """
        reassign_worker_ops = [
            chunk.op for chunk in self._chunk_graph if chunk.op.reassign_worker
        ]
        start_ops = (
            list(self._iter_start_ops(self._chunk_graph))
            if len(self._chunk_graph) > 0
            else []
        )

        # assign start chunks
        to_assign_ops = start_ops + reassign_worker_ops
        assigner = self._graph_assigner_cls(
            self._chunk_graph, to_assign_ops, self._band_slots
        )
        # assign expect workers
        cur_assigns = {
            op.key: self._to_band(op.expect_worker)
            for op in start_ops
            if op.expect_worker is not None
        }
        logger.debug(
            "Start to assign %s start chunks for task %s",
            len(start_ops),
            self._task.task_id,
        )
        chunk_to_bands = assigner.assign(cur_assigns=cur_assigns)
        logger.debug(
            "Assigned %s start chunks for task %s", len(start_ops), self._task.task_id
        )
        # assign expect workers for those specified with `expect_worker`
        # skip `start_ops`, which have been assigned before
        for chunk in self._chunk_graph:
            if chunk not in start_ops and chunk.op.expect_worker is not None:
                chunk_to_bands[chunk] = self._to_band(chunk.op.expect_worker)

        # color nodes
        if self._fuse_enabled:
            logger.debug("Start to fuse chunks for task %s", self._task.task_id)
            # sort start chunks in coloring as start_ops
            op_key_to_chunks = defaultdict(list)
            for chunk in self._chunk_graph:
                op_key_to_chunks[chunk.op.key].append(chunk)
            init_chunk_to_bands = dict()
            for start_op in start_ops:
                for start_chunk in op_key_to_chunks[start_op.key]:
                    init_chunk_to_bands[start_chunk] = chunk_to_bands[start_chunk]
            coloring = Coloring(
                self._chunk_graph,
                list(self._band_slots),
                init_chunk_to_bands,
                initial_same_color_num=getattr(
                    self._config, "initial_same_color_num", None
                ),
                as_broadcaster_successor_num=getattr(
                    self._config, "as_broadcaster_successor_num", None
                ),
            )
            chunk_to_colors = coloring.color()
        else:
            # if not fuse enabled, color all chunks with different colors
            chunk_to_colors = {
                c: i for i, c in enumerate(self._chunk_graph.topological_iter())
            }
        color_to_chunks = defaultdict(list)
        for chunk, color in chunk_to_colors.items():
            color_to_chunks[color].append(chunk)

        # gen subtask graph
        subtask_graph = SubtaskGraph()
        chunk_to_fetch_chunk = dict()
        chunk_to_subtask = dict()
        # states
        visited = set()
        logic_key_to_subtasks = defaultdict(list)
        for chunk in self._chunk_graph.topological_iter():
            if chunk in visited:
                continue

            color = chunk_to_colors[chunk]
            same_color_chunks = color_to_chunks[color]
            if all(isinstance(c.op, Fetch) for c in same_color_chunks):
                # all fetch ops, no need to gen subtask
                continue
            subtask, inp_subtasks = self._gen_subtask_info(
                same_color_chunks,
                chunk_to_subtask,
                chunk_to_bands,
                chunk_to_fetch_chunk,
            )
            subtask_graph.add_node(subtask)
            logic_key_to_subtasks[subtask.logic_key].append(subtask)
            for inp_subtask in inp_subtasks:
                subtask_graph.add_edge(inp_subtask, subtask)

            for c in same_color_chunks:
                chunk_to_subtask[c] = subtask
            visited.update(same_color_chunks)

        for subtasks in logic_key_to_subtasks.values():
            for logic_index, subtask in enumerate(subtasks):
                subtask.logic_index = logic_index
                subtask.logic_parallelism = len(subtasks)
        return subtask_graph



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import logging
from functools import partial
from typing import Callable, Dict, List, Iterable, Set, Tuple

from ....config import Config
from ....core import TileableGraph, ChunkGraph, ChunkGraphBuilder
from ....core.graph.builder.chunk import Tiler, tile_gen_type
from ....core.operand import Fetch
from ....typing import TileableType, ChunkType
from ....optimization.logical.chunk import optimize as optimize_chunk_graph
from ....optimization.logical.tileable import optimize as optimize_tileable_graph
from ....typing import BandType
from ...subtask import SubtaskGraph
from ..analyzer import GraphAnalyzer
from ..core import Task

logger = logging.getLogger(__name__)


class CancellableTiler(Tiler):
    def __init__(
        self,
        tileable_graph: TileableGraph,
        tile_context: Dict[TileableType, TileableType],
        processed_chunks: Set[ChunkType],
        add_nodes: Callable,
        cancelled: asyncio.Event = None,
    ):
        super().__init__(tileable_graph, tile_context, processed_chunks, add_nodes)
        self._cancelled = cancelled

    @property
    def cancelled(self):
        return self._cancelled.is_set()

    def _gen_tileable_handlers(
        self, next_tileable_handlers: List[Tuple[TileableType, tile_gen_type]]
    ):
        for tileable, tile_handler in super()._gen_tileable_handlers(
            next_tileable_handlers
        ):
            if not self.cancelled:
                yield tileable, tile_handler
            else:
                break

    def _gen_result_chunks(
        self,
        chunk_graph: ChunkGraph,
        next_tileable_handlers: List[Tuple[TileableType, tile_gen_type]],
    ):
        if not self.cancelled:
            return super()._gen_result_chunks(chunk_graph, next_tileable_handlers)
        else:
            return

    def __iter__(self):
        while self._tileable_handlers:
            to_update_tileables = self._iter()
            if not self.cancelled:
                yield self._cur_chunk_graph
            if not self.cancelled:
                for t in to_update_tileables:
                    t.refresh_params()
            else:
                break


class TaskPreprocessor:
    __slots__ = (
        "_task",
        "tileable_graph",
        "tile_context",
        "_config",
        "tileable_optimization_records",
        "chunk_optimization_records_list",
        "_cancelled",
        "_done",
    )

    tile_context: Dict[TileableType, TileableType]

    def __init__(
        self,
        task: Task,
        tiled_context: Dict[TileableType, TileableType] = None,
        config: Config = None,
    ):
        self._task = task
        self.tileable_graph = task.tileable_graph
        self._config = config

        self.tile_context = tiled_context
        self.tileable_optimization_records = None
        self.chunk_optimization_records_list = []

        self._cancelled = asyncio.Event()
        self._done = asyncio.Event()

    def optimize(self) -> TileableGraph:
        """
        Optimize tileable graph.

        Returns
        -------
        optimized_graph: TileableGraph

        """
        if self._config.optimize_tileable_graph:
            # enable optimization
            self.tileable_optimization_records = optimize_tileable_graph(
                self.tileable_graph
            )
        return self.tileable_graph

    def _fill_fetch_tileable_with_chunks(self, tileable_graph: TileableGraph):
        for t in tileable_graph:
            if isinstance(t.op, Fetch) and t in self.tile_context:
                tiled = self.tile_context[t]
                t._chunks = tiled.chunks
                t._nsplits = tiled.nsplits

    def tile(self, tileable_graph: TileableGraph) -> Iterable[ChunkGraph]:
        """
        Generate chunk graphs

        Returns
        -------
        chunk_graph_generator: Generator
             Chunk graphs.
        """
        self._fill_fetch_tileable_with_chunks(tileable_graph)
        # iterative chunk graph builder
        chunk_graph_builder = ChunkGraphBuilder(
            tileable_graph,
            fuse_enabled=self._task.fuse_enabled,
            tile_context=self.tile_context,
            tiler_cls=partial(CancellableTiler, cancelled=self._cancelled),
        )
        optimize = self._config.optimize_chunk_graph
        meta_updated = set()
        for chunk_graph in chunk_graph_builder.build():
            # optimize chunk graph
            if optimize:
                self.chunk_optimization_records_list.append(
                    optimize_chunk_graph(chunk_graph)
                )
            yield chunk_graph
            # update tileables' meta
            self._update_tileables_params(tileable_graph, meta_updated)

    def analyze(
        self,
        chunk_graph: ChunkGraph,
        available_bands: Dict[BandType, int],
        stage_id: str = None,
    ) -> SubtaskGraph:
        logger.debug("Start to gen subtask graph for task %s", self._task.task_id)
        task = self._task
        analyzer = GraphAnalyzer(
            chunk_graph, available_bands, task, self._config, stage_id=stage_id
        )
        graph = analyzer.gen_subtask_graph()
        logger.debug(
            "Generated subtask graph of %s subtasks for task %s",
            len(graph),
            self._task.task_id,
        )
        return graph

    def _get_done(self):
        return self._done.is_set()

    def _set_done(self, is_done: bool):
        if is_done:
            self._done.set()
        else:  # pragma: no cover
            self._done.clear()

    done = property(_get_done, _set_done)

    def cancel(self):
        self._cancelled.set()

    def get_tiled(self, tileable: TileableType):
        tileable = tileable.data if hasattr(tileable, "data") else tileable
        return self.tile_context[tileable]

    def _update_tileable_params(
        self, tileable: TileableType, tiled: TileableType  # pylint: disable=no-self-use
    ):
        tiled.refresh_params()
        tileable.params = tiled.params

    def _update_tileables_params(
        self, tileable_graph: TileableGraph, updated: Set[TileableType]
    ):
        for tileable in tileable_graph:
            if tileable in updated:
                continue
            tiled_tileable = self.tile_context.get(tileable)
            if tiled_tileable is not None:
                self._update_tileable_params(tileable, tiled_tileable)
                updated.add(tileable)

    def __await__(self):
        return self._done.wait().__await__()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures as futures
import itertools
import logging
import multiprocessing
import os
import threading
from abc import ABC, ABCMeta, abstractmethod
from typing import Dict, List, Type, TypeVar, Coroutine, Callable, Union, Optional

from ...core.entrypoints import init_extension_entrypoints
from ...metrics import init_metrics
from ...utils import implements, to_binary
from ...utils import lazy_import, register_asyncio_task_timeout_detector
from ..api import Actor
from ..core import ActorRef, register_local_pool
from ..debug import record_message_trace, debug_async_timeout
from ..errors import ActorAlreadyExist, ActorNotExist, ServerClosed, CannotCancelTask
from ..utils import create_actor_ref
from .allocate_strategy import allocated_type, AddressSpecified
from .communication import Channel, Server, get_server_type, gen_local_address
from .communication.errors import ChannelClosed
from .config import ActorPoolConfig
from .core import result_message_type, ActorCaller
from .message import (
    _MessageBase,
    new_message_id,
    DEFAULT_PROTOCOL,
    MessageType,
    ResultMessage,
    ErrorMessage,
    CreateActorMessage,
    HasActorMessage,
    DestroyActorMessage,
    ActorRefMessage,
    SendMessage,
    TellMessage,
    CancelMessage,
    ControlMessage,
    ControlMessageType,
)
from .router import Router

logger = logging.getLogger(__name__)
ray = lazy_import("ray")


class _ErrorProcessor:
    def __init__(self, address: str, message_id: bytes, protocol):
        self._address = address
        self._message_id = message_id
        self._protocol = protocol
        self.result = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.result is None:
            self.result = ErrorMessage(
                self._message_id,
                self._address,
                os.getpid(),
                exc_type,
                exc_val,
                exc_tb,
                protocol=self._protocol,
            )
            return True


def _register_message_handler(pool_type: Type["AbstractActorPool"]):
    pool_type._message_handler = dict()
    for message_type, handler in [
        (MessageType.create_actor, pool_type.create_actor),
        (MessageType.destroy_actor, pool_type.destroy_actor),
        (MessageType.has_actor, pool_type.has_actor),
        (MessageType.actor_ref, pool_type.actor_ref),
        (MessageType.send, pool_type.send),
        (MessageType.tell, pool_type.tell),
        (MessageType.cancel, pool_type.cancel),
        (MessageType.control, pool_type.handle_control_command),
    ]:
        pool_type._message_handler[message_type] = handler
    return pool_type


class AbstractActorPool(ABC):
    __slots__ = (
        "process_index",
        "label",
        "external_address",
        "internal_address",
        "env",
        "_servers",
        "_router",
        "_config",
        "_stopped",
        "_actors",
        "_caller",
        "_process_messages",
        "_asyncio_task_timeout_detector_task",
    )

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: Dict,
        router: Router,
        config: ActorPoolConfig,
        servers: List[Server],
    ):
        # register local pool for local actor lookup.
        # The pool is weakrefed, so we don't need to unregister it.
        register_local_pool(external_address, self)
        self.process_index = process_index
        self.label = label
        self.external_address = external_address
        self.internal_address = internal_address
        self.env = env
        self._router = router
        self._config = config
        self._servers = servers

        self._stopped = asyncio.Event()

        # states
        # actor id -> actor
        self._actors: Dict[bytes, Actor] = dict()
        # message id -> future
        self._process_messages: Dict[bytes, asyncio.Future] = dict()

        # manage async actor callers
        self._caller = ActorCaller()
        self._asyncio_task_timeout_detector_task = (
            register_asyncio_task_timeout_detector()
        )
        # load third party extensions.
        init_extension_entrypoints()
        # init metrics
        metric_configs = self._config.get_metric_configs()
        init_metrics(metric_configs.get("backend"), port=metric_configs.get("port"))

    @property
    def router(self):
        return self._router

    @abstractmethod
    async def create_actor(self, message: CreateActorMessage) -> result_message_type:
        """
        Create an actor.

        Parameters
        ----------
        message: CreateActorMessage
            message to create an actor.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        """
        Check if an actor exists or not.

        Parameters
        ----------
        message: HasActorMessage
            message

        Returns
        -------
        result_message
            result message contains if an actor exists or not.
        """

    @abstractmethod
    async def destroy_actor(self, message: DestroyActorMessage) -> result_message_type:
        """
        Destroy an actor.

        Parameters
        ----------
        message: DestroyActorMessage
            message to destroy an actor.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def actor_ref(self, message: ActorRefMessage) -> result_message_type:
        """
        Get an actor's ref.

        Parameters
        ----------
        message: ActorRefMessage
            message to get an actor's ref.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def send(self, message: SendMessage) -> result_message_type:
        """
        Send a message to some actor.

        Parameters
        ----------
        message: SendMessage
            Message to send.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def tell(self, message: TellMessage) -> result_message_type:
        """
        Tell message to some actor.

        Parameters
        ----------
        message: TellMessage
            Message to tell.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def cancel(self, message: CancelMessage) -> result_message_type:
        """
        Cancel message that sent

        Parameters
        ----------
        message: CancelMessage
            Cancel message.

        Returns
        -------
        result_message
            result or error message
        """

    async def handle_control_command(
        self, message: ControlMessage
    ) -> result_message_type:
        """
        Handle control command.

        Parameters
        ----------
        message: ControlMessage
            Control message.

        Returns
        -------
        result_message
            result or error message.
        """
        with _ErrorProcessor(
            self.external_address, message.message_id, protocol=message.protocol
        ) as processor:
            content = True
            if message.control_message_type == ControlMessageType.stop:
                await self.stop()
            elif message.control_message_type == ControlMessageType.sync_config:
                actor_pool_config: ActorPoolConfig = message.content
                self._config = actor_pool_config
                # remove router from global one
                global_router = Router.get_instance()
                global_router.remove_router(self._router)
                # update router
                self._router.set_mapping(
                    actor_pool_config.external_to_internal_address_map
                )
                # update global router
                global_router.add_router(self._router)
            elif message.control_message_type == ControlMessageType.get_config:
                if message.content == "main_pool_address":
                    main_process_index = self._config.get_process_indexes()[0]
                    content = self._config.get_pool_config(main_process_index)[
                        "external_address"
                    ][0]
                else:
                    content = self._config
            else:  # pragma: no cover
                raise TypeError(
                    f"Unable to handle control message "
                    f"with type {message.control_message_type}"
                )
            processor.result = ResultMessage(
                message.message_id, content, protocol=message.protocol
            )

        return processor.result

    async def _run_coro(self, message_id: bytes, coro: Coroutine):
        self._process_messages[message_id] = asyncio.tasks.current_task()
        try:
            return await coro
        finally:
            self._process_messages.pop(message_id, None)

    async def process_message(self, message: _MessageBase, channel: Channel):
        handler = self._message_handler[message.message_type]
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            with debug_async_timeout(
                "process_message_timeout",
                "Process message %s of channel %s timeout.",
                message,
                channel,
            ):
                processor.result = await self._run_coro(
                    message.message_id, handler(self, message)
                )
        try:
            await channel.send(processor.result)
        except (ChannelClosed, ConnectionResetError):
            if not self._stopped.is_set():
                raise

    async def call(
        self, dest_address: str, message: _MessageBase
    ) -> result_message_type:
        return await self._caller.call(self._router, dest_address, message)

    @staticmethod
    def _parse_config(config: Dict, kw: Dict) -> Dict:
        actor_pool_config: ActorPoolConfig = config.pop("actor_pool_config")
        kw["config"] = actor_pool_config
        kw["process_index"] = process_index = config.pop("process_index")
        curr_pool_config = actor_pool_config.get_pool_config(process_index)
        kw["label"] = curr_pool_config["label"]
        external_addresses = curr_pool_config["external_address"]
        kw["external_address"] = external_addresses[0]
        kw["internal_address"] = curr_pool_config["internal_address"]
        kw["router"] = Router(
            external_addresses,
            gen_local_address(process_index),
            actor_pool_config.external_to_internal_address_map,
        )
        kw["env"] = curr_pool_config["env"]

        if config:  # pragma: no cover
            raise TypeError(
                f"Creating pool got unexpected " f'arguments: {",".join(config)}'
            )

        return kw

    @classmethod
    @abstractmethod
    async def create(cls, config: Dict) -> "AbstractActorPool":
        """
        Create an actor pool.

        Parameters
        ----------
        config: Dict
            configurations.

        Returns
        -------
        actor_pool:
            Actor pool.
        """

    async def start(self):
        if self._stopped.is_set():
            raise RuntimeError("pool has been stopped, cannot start again")
        start_servers = [server.start() for server in self._servers]
        await asyncio.gather(*start_servers)

    async def join(self, timeout: float = None):
        wait_stopped = asyncio.create_task(self._stopped.wait())

        try:
            await asyncio.wait_for(wait_stopped, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):  # pragma: no cover
            wait_stopped.cancel()

    async def stop(self):
        try:
            # clean global router
            Router.get_instance().remove_router(self._router)
            stop_tasks = []
            # stop all servers
            stop_tasks.extend([server.stop() for server in self._servers])
            # stop all clients
            stop_tasks.append(self._caller.stop())
            await asyncio.gather(*stop_tasks)

            self._servers = []
            if self._asyncio_task_timeout_detector_task:  # pragma: no cover
                self._asyncio_task_timeout_detector_task.cancel()
        finally:
            self._stopped.set()

    @property
    def stopped(self) -> bool:
        return self._stopped.is_set()

    async def on_new_channel(self, channel: Channel):
        while not self._stopped.is_set():
            try:
                message = await channel.recv()
            except EOFError:
                # no data to read, check channel
                try:
                    await channel.close()
                except (ConnectionError, EOFError):
                    # close failed, ignore
                    pass
                return
            asyncio.create_task(self.process_message(message, channel))
            # delete to release the reference of message
            del message
            await asyncio.sleep(0)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class ActorPoolBase(AbstractActorPool, metaclass=ABCMeta):
    __slots__ = ()

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_id
            if actor_id in self._actors:
                raise ActorAlreadyExist(
                    f"Actor {actor_id} already exist, " f"cannot create"
                )

            actor = message.actor_cls(*message.args, **message.kwargs)
            actor.uid = actor_id
            actor.address = address = self.external_address
            self._actors[actor_id] = actor
            await self._run_coro(message.message_id, actor.__post_create__())

            result = ActorRef(address, actor_id)
            # ensemble result message
            processor.result = ResultMessage(
                message.message_id, result, protocol=message.protocol
            )
        return processor.result

    @implements(AbstractActorPool.has_actor)
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        result = ResultMessage(
            message.message_id,
            message.actor_ref.uid in self._actors,
            protocol=message.protocol,
        )
        return result

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            try:
                actor = self._actors[actor_id]
            except KeyError:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            await self._run_coro(message.message_id, actor.__pre_destroy__())
            del self._actors[actor_id]

            processor.result = ResultMessage(
                message.message_id, actor_id, protocol=message.protocol
            )
        return processor.result

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            result = ResultMessage(
                message.message_id,
                ActorRef(self.external_address, actor_id),
                protocol=message.protocol,
            )
            processor.result = result
        return processor.result

    @implements(AbstractActorPool.send)
    async def send(self, message: SendMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor, record_message_trace(message):
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            coro = self._actors[actor_id].__on_receive__(message.content)
            result = await self._run_coro(message.message_id, coro)
            processor.result = ResultMessage(
                message.message_id,
                result,
                protocol=message.protocol,
                profiling_context=message.profiling_context,
            )
        return processor.result

    @implements(AbstractActorPool.tell)
    async def tell(self, message: TellMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:  # pragma: no cover
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            call = self._actors[actor_id].__on_receive__(message.content)
            # asynchronously run, tell does not care about result
            asyncio.create_task(call)
            await asyncio.sleep(0)
            processor.result = ResultMessage(
                message.message_id,
                None,
                protocol=message.protocol,
                profiling_context=message.profiling_context,
            )
        return processor.result

    @implements(AbstractActorPool.cancel)
    async def cancel(self, message: CancelMessage) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            future = self._process_messages.get(message.cancel_message_id)
            if future is None or future.done():  # pragma: no cover
                raise CannotCancelTask(
                    "Task not exists, maybe it is done or cancelled already"
                )
            future.cancel()
            processor.result = ResultMessage(
                message.message_id, True, protocol=message.protocol
            )
        return processor.result

    @staticmethod
    def _set_global_router(router: Router):
        # be cautious about setting global router
        # for instance, multiple main pool may be created in the same process

        # get default router or create an empty one
        default_router = Router.get_instance_or_empty()
        Router.set_instance(default_router)
        # append this router to global
        default_router.add_router(router)

    @classmethod
    @implements(AbstractActorPool.create)
    async def create(cls, config: Dict) -> "ActorPoolType":
        config = config.copy()
        kw = dict()
        cls._parse_config(config, kw)
        process_index: int = kw["process_index"]
        actor_pool_config = kw["config"]
        external_addresses = actor_pool_config.get_pool_config(process_index)[
            "external_address"
        ]
        internal_address = kw["internal_address"]

        # import predefined modules
        modules = actor_pool_config.get_pool_config(process_index)["modules"] or []
        for mod in modules:
            __import__(mod, globals(), locals(), [])

        # set default router
        # actor context would be able to use exact client
        cls._set_global_router(kw["router"])

        def handle_channel(channel):
            return pool.on_new_channel(channel)

        # create servers
        create_server_tasks = []
        for addr in set(
            external_addresses + [internal_address, gen_local_address(process_index)]
        ):
            server_type = get_server_type(addr)
            task = asyncio.create_task(
                server_type.create(dict(address=addr, handle_channel=handle_channel))
            )
            create_server_tasks.append(task)
        await asyncio.gather(*create_server_tasks)
        kw["servers"] = [f.result() for f in create_server_tasks]

        # create pool
        pool = cls(**kw)
        return pool


ActorPoolType = TypeVar("ActorPoolType", bound=AbstractActorPool)
MainActorPoolType = TypeVar("MainActorPoolType", bound="MainActorPoolBase")
SubProcessHandle = Union[multiprocessing.Process, "ray.actor.ActorHandle"]


class SubActorPoolBase(ActorPoolBase):
    __slots__ = ("_main_address",)

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: Dict,
        router: Router,
        config: ActorPoolConfig,
        servers: List[Server],
        main_address: str,
    ):
        super().__init__(
            process_index,
            label,
            external_address,
            internal_address,
            env,
            router,
            config,
            servers,
        )
        self._main_address = main_address

    async def notify_main_pool_to_destroy(
        self, message: DestroyActorMessage
    ):  # pragma: no cover
        await self.call(self._main_address, message)

    async def notify_main_pool_to_create(self, message: CreateActorMessage):
        reg_message = ControlMessage(
            new_message_id(),
            self.external_address,
            ControlMessageType.add_sub_pool_actor,
            (self.external_address, message.allocate_strategy, message),
        )
        await self.call(self._main_address, reg_message)

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> result_message_type:
        result = await super().create_actor(message)
        if not message.from_main:
            await self.notify_main_pool_to_create(message)
        return result

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> result_message_type:
        result = await super().actor_ref(message)
        if isinstance(result, ErrorMessage):
            message.actor_ref.address = self._main_address
            result = await self.call(self._main_address, message)
        return result

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> result_message_type:
        result = await super().destroy_actor(message)
        if isinstance(result, ResultMessage) and not message.from_main:
            # sync back to main actor pool
            await self.notify_main_pool_to_destroy(message)
        return result

    @staticmethod
    def _parse_config(config: Dict, kw: Dict) -> Dict:
        kw = AbstractActorPool._parse_config(config, kw)
        config: ActorPoolConfig = kw["config"]
        main_process_index = config.get_process_indexes()[0]
        kw["main_address"] = config.get_pool_config(main_process_index)[
            "external_address"
        ][0]
        return kw


class MainActorPoolBase(ActorPoolBase):
    __slots__ = (
        "_allocated_actors",
        "sub_actor_pool_manager",
        "_auto_recover",
        "_monitor_task",
        "_on_process_down",
        "_on_process_recover",
        "_recover_events",
    )

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: Dict,
        router: Router,
        config: ActorPoolConfig,
        servers: List[Server],
        subprocess_start_method: str = None,
        auto_recover: Union[str, bool] = "actor",
        on_process_down: Callable[[MainActorPoolType, str], None] = None,
        on_process_recover: Callable[[MainActorPoolType, str], None] = None,
    ):
        super().__init__(
            process_index,
            label,
            external_address,
            internal_address,
            env,
            router,
            config,
            servers,
        )
        self._subprocess_start_method = subprocess_start_method

        # auto recovering
        self._auto_recover = auto_recover
        self._monitor_task: Optional[asyncio.Task] = None
        self._on_process_down = on_process_down
        self._on_process_recover = on_process_recover
        self._recover_events: Dict[str, asyncio.Event] = dict()

        # states
        self._allocated_actors: allocated_type = {
            addr: dict() for addr in self._config.get_external_addresses()
        }
        self._allocation_lock = threading.Lock()

        self.sub_processes: Dict[str, SubProcessHandle] = dict()

    _process_index_gen = itertools.count()

    @classmethod
    def process_index_gen(cls, address):
        # make sure different processes does not share process indexes
        pid = os.getpid()
        for idx in cls._process_index_gen:
            yield pid << 16 + idx

    @property
    def _sub_processes(self):
        return self.sub_processes

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> result_message_type:
        with _ErrorProcessor(
            address=self.external_address,
            message_id=message.message_id,
            protocol=message.protocol,
        ) as processor:
            allocate_strategy = message.allocate_strategy
            with self._allocation_lock:
                # get allocated address according to corresponding strategy
                address = allocate_strategy.get_allocated_address(
                    self._config, self._allocated_actors
                )
                # set placeholder to make sure this label is occupied
                self._allocated_actors[address][None] = (allocate_strategy, message)
            if address == self.external_address:
                # creating actor on main actor pool
                result = await super().create_actor(message)
                if isinstance(result, ResultMessage):
                    self._allocated_actors[self.external_address][result.result] = (
                        allocate_strategy,
                        message,
                    )
                processor.result = result
            else:
                # creating actor on sub actor pool
                # rewrite allocate strategy to AddressSpecified
                new_allocate_strategy = AddressSpecified(address)
                new_create_actor_message = CreateActorMessage(
                    message.message_id,
                    message.actor_cls,
                    message.actor_id,
                    message.args,
                    message.kwargs,
                    allocate_strategy=new_allocate_strategy,
                    from_main=True,
                    protocol=message.protocol,
                    message_trace=message.message_trace,
                )
                result = await self.call(address, new_create_actor_message)
                if isinstance(result, ResultMessage):
                    self._allocated_actors[address][result.result] = (
                        allocate_strategy,
                        new_create_actor_message,
                    )
                processor.result = result

            # revert placeholder
            self._allocated_actors[address].pop(None, None)

        return processor.result

    @implements(AbstractActorPool.has_actor)
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        actor_ref = message.actor_ref
        # lookup allocated
        for address, item in self._allocated_actors.items():
            ref = create_actor_ref(address, to_binary(actor_ref.uid))
            if ref in item:
                return ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )

        return ResultMessage(message.message_id, False, protocol=message.protocol)

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> result_message_type:
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        real_actor_ref = result.result
        if real_actor_ref.address == self.external_address:
            await super().destroy_actor(message)
            del self._allocated_actors[self.external_address][real_actor_ref]
            return ResultMessage(
                message.message_id, real_actor_ref.uid, protocol=message.protocol
            )
        # remove allocated actor ref
        self._allocated_actors[real_actor_ref.address].pop(real_actor_ref, None)
        new_destroy_message = DestroyActorMessage(
            message.message_id,
            real_actor_ref,
            from_main=True,
            protocol=message.protocol,
        )
        return await self.call(real_actor_ref.address, new_destroy_message)

    @implements(AbstractActorPool.send)
    async def send(self, message: SendMessage) -> result_message_type:
        if message.actor_ref.uid in self._actors:
            return await super().send(message)
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        actor_ref = result.result
        new_send_message = SendMessage(
            message.message_id,
            actor_ref,
            message.content,
            protocol=message.protocol,
            message_trace=message.message_trace,
        )
        return await self.call(actor_ref.address, new_send_message)

    @implements(AbstractActorPool.tell)
    async def tell(self, message: TellMessage) -> result_message_type:
        if message.actor_ref.uid in self._actors:
            return await super().tell(message)
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        actor_ref = result.result
        new_tell_message = TellMessage(
            message.message_id,
            actor_ref,
            message.content,
            protocol=message.protocol,
            message_trace=message.message_trace,
        )
        return await self.call(actor_ref.address, new_tell_message)

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> result_message_type:
        actor_ref = message.actor_ref
        actor_ref.uid = to_binary(actor_ref.uid)
        if actor_ref.address == self.external_address and actor_ref.uid in self._actors:
            return ResultMessage(
                message.message_id, actor_ref, protocol=message.protocol
            )

        # lookup allocated
        for address, item in self._allocated_actors.items():
            ref = create_actor_ref(address, actor_ref.uid)
            if ref in item:
                return ResultMessage(message.message_id, ref, protocol=message.protocol)

        with _ErrorProcessor(
            self.external_address, message.message_id, protocol=message.protocol
        ) as processor:
            raise ActorNotExist(
                f"Actor {actor_ref.uid} does not exist in {actor_ref.address}"
            )

        return processor.result

    @implements(AbstractActorPool.cancel)
    async def cancel(self, message: CancelMessage) -> result_message_type:
        if message.address == self.external_address:
            # local message
            return await super().cancel(message)
        # redirect to sub pool
        return await self.call(message.address, message)

    @implements(AbstractActorPool.handle_control_command)
    async def handle_control_command(
        self, message: ControlMessage
    ) -> result_message_type:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            if message.address == self.external_address:
                if message.control_message_type == ControlMessageType.sync_config:
                    # sync config, need to notify all sub pools
                    tasks = []
                    for addr in self.sub_processes:
                        control_message = ControlMessage(
                            new_message_id(),
                            addr,
                            message.control_message_type,
                            message.content,
                            protocol=message.protocol,
                            message_trace=message.message_trace,
                        )
                        tasks.append(
                            asyncio.create_task(self.call(addr, control_message))
                        )
                    # call super
                    task = asyncio.create_task(super().handle_control_command(message))
                    tasks.append(task)
                    await asyncio.gather(*tasks)
                    processor.result = await task
                else:
                    processor.result = await super().handle_control_command(message)
            elif message.control_message_type == ControlMessageType.stop:
                timeout, force = (
                    message.content if message.content is not None else (None, False)
                )
                await self.stop_sub_pool(
                    message.address,
                    self.sub_processes[message.address],
                    timeout=timeout,
                    force=force,
                )
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            elif message.control_message_type == ControlMessageType.wait_pool_recovered:
                if self._auto_recover and message.address not in self._recover_events:
                    self._recover_events[message.address] = asyncio.Event()

                event = self._recover_events.get(message.address, None)
                if event is not None:
                    await event.wait()
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            elif message.control_message_type == ControlMessageType.add_sub_pool_actor:
                address, allocate_strategy, create_message = message.content
                create_message.from_main = True
                ref = create_actor_ref(address, to_binary(create_message.actor_id))
                self._allocated_actors[address][ref] = (
                    allocate_strategy,
                    create_message,
                )
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            else:
                processor.result = await self.call(message.address, message)
        return processor.result

    @staticmethod
    def _parse_config(config: Dict, kw: Dict) -> Dict:
        kw["subprocess_start_method"] = config.pop("start_method", None)
        kw["auto_recover"] = config.pop("auto_recover", "actor")
        kw["on_process_down"] = config.pop("on_process_down", None)
        kw["on_process_recover"] = config.pop("on_process_recover", None)
        kw = AbstractActorPool._parse_config(config, kw)
        return kw

    @classmethod
    @implements(AbstractActorPool.create)
    async def create(cls, config: Dict) -> MainActorPoolType:
        config = config.copy()
        actor_pool_config: ActorPoolConfig = config.get("actor_pool_config")
        start_method = config.get("start_method", None)
        if "process_index" not in config:
            config["process_index"] = actor_pool_config.get_process_indexes()[0]
        curr_process_index = config.get("process_index")

        tasks = []
        # create sub actor pools
        n_sub_pool = actor_pool_config.n_pool - 1
        if n_sub_pool > 0:
            process_indexes = actor_pool_config.get_process_indexes()
            for process_index in process_indexes:
                if process_index == curr_process_index:
                    continue
                create_pool_task = asyncio.create_task(
                    cls.start_sub_pool(actor_pool_config, process_index, start_method)
                )
                await asyncio.sleep(0)
                # await create_pool_task
                tasks.append(create_pool_task)

        processes = await cls.wait_sub_pools_ready(tasks)
        # create main actor pool
        pool: MainActorPoolType = await super().create(config)
        addresses = actor_pool_config.get_external_addresses()[1:]

        assert len(addresses) == len(
            processes
        ), f"addresses {addresses}, processes {processes}"
        for addr, proc in zip(addresses, processes):
            pool.attach_sub_process(addr, proc)
        return pool

    async def start_monitor(self):
        if self._monitor_task is None:
            self._monitor_task = asyncio.create_task(self.monitor_sub_pools())
        return self._monitor_task

    @implements(AbstractActorPool.stop)
    async def stop(self):
        # turn off auto recover to avoid errors
        self._auto_recover = False
        self._stopped.set()
        if self._monitor_task and not self._monitor_task.done():
            await self._monitor_task
            self._monitor_task = None
        await self.stop_sub_pools()
        await super().stop()

    @classmethod
    @abstractmethod
    async def start_sub_pool(
        cls,
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_method: str = None,
    ):
        """Start a sub actor pool"""

    @classmethod
    @abstractmethod
    async def wait_sub_pools_ready(cls, create_pool_tasks: List[asyncio.Task]):
        """Wait all sub pools ready"""

    def attach_sub_process(self, external_address: str, process: SubProcessHandle):
        self.sub_processes[external_address] = process

    async def stop_sub_pools(self):
        to_stop_processes: Dict[str, SubProcessHandle] = dict()
        for address, process in self.sub_processes.items():
            if not await self.is_sub_pool_alive(process):
                continue
            to_stop_processes[address] = process

        tasks = []
        for address, process in to_stop_processes.items():
            tasks.append(self.stop_sub_pool(address, process))
        await asyncio.gather(*tasks)

    async def stop_sub_pool(
        self,
        address: str,
        process: SubProcessHandle,
        timeout: float = None,
        force: bool = False,
    ):
        if force:
            await self.kill_sub_pool(process, force=True)
            return

        stop_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.stop,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        try:
            if timeout is None:
                message = await self.call(address, stop_message)
                if isinstance(message, ErrorMessage):  # pragma: no cover
                    raise message.as_instanceof_cause()
            else:
                call = asyncio.create_task(self.call(address, stop_message))
                try:
                    await asyncio.wait_for(call, timeout)
                except (futures.TimeoutError, asyncio.TimeoutError):  # pragma: no cover
                    # timeout, just let kill to finish it
                    force = True
        except (ConnectionError, ServerClosed):  # pragma: no cover
            # process dead maybe, ignore it
            pass
        # kill process
        await self.kill_sub_pool(process, force=force)

    @abstractmethod
    async def kill_sub_pool(self, process: SubProcessHandle, force: bool = False):
        """Kill a sub actor pool"""

    @abstractmethod
    async def is_sub_pool_alive(self, process: SubProcessHandle):
        """
        Check whether sub pool process is alive
        Parameters
        ----------
        process : SubProcessHandle
            sub pool process handle
        Returns
        -------
        bool
        """

    @abstractmethod
    def recover_sub_pool(self, address):
        """Recover a sub actor pool"""

    def process_sub_pool_lost(self, address: str):
        if self._auto_recover in (False, "process"):
            # process down, when not auto_recover
            # or only recover process, remove all created actors
            self._allocated_actors[address] = dict()

    async def monitor_sub_pools(self):
        try:
            while not self._stopped.is_set():
                for address, process in self.sub_processes.items():
                    try:
                        recover_events_discovered = address in self._recover_events
                        if not await self.is_sub_pool_alive(
                            process
                        ):  # pragma: no cover
                            if self._on_process_down is not None:
                                self._on_process_down(self, address)
                            self.process_sub_pool_lost(address)
                            if self._auto_recover:
                                await self.recover_sub_pool(address)
                                if self._on_process_recover is not None:
                                    self._on_process_recover(self, address)
                        if recover_events_discovered:
                            event = self._recover_events.pop(address)
                            event.set()
                    except asyncio.CancelledError:
                        raise
                    except Exception:
                        # log the exception instead of stop monitoring the
                        # sub pool silently.
                        logger.exception("Monitor sub pool %s failed", address)

                # check every half second
                await asyncio.sleep(0.5)
        except asyncio.CancelledError:  # pragma: no cover
            # cancelled
            return

    @classmethod
    @abstractmethod
    def get_external_addresses(
        cls, address: str, n_process: int = None, ports: List[int] = None
    ):
        """Returns external addresses for n pool processes"""

    @classmethod
    @abstractmethod
    def gen_internal_address(
        cls, process_index: int, external_address: str = None
    ) -> str:
        """Returns internal address for pool of specified process index"""


async def create_actor_pool(
    address: str,
    *,
    pool_cls: Type[MainActorPoolType] = None,
    n_process: int = None,
    labels: List[str] = None,
    ports: List[int] = None,
    envs: List[Dict] = None,
    subprocess_start_method: str = None,
    auto_recover: Union[str, bool] = "actor",
    modules: List[str] = None,
    suspend_sigint: bool = None,
    use_uvloop: Union[str, bool] = "auto",
    logging_conf: Union[Dict, None] = None,
    on_process_down: Callable[[MainActorPoolType, str], None] = None,
    on_process_recover: Callable[[MainActorPoolType, str], None] = None,
    **kwargs,
) -> MainActorPoolType:
    if n_process is None:
        n_process = multiprocessing.cpu_count()
    if labels and len(labels) != n_process + 1:
        raise ValueError(
            f"`labels` should be of size {n_process + 1}, " f"got {len(labels)}"
        )
    if envs and len(envs) != n_process:
        raise ValueError(f"`envs` should be of size {n_process}, " f"got {len(envs)}")
    if auto_recover is True:
        auto_recover = "actor"
    if auto_recover not in ("actor", "process", False):
        raise ValueError(
            f'`auto_recover` should be one of "actor", "process", '
            f"True or False, got {auto_recover}"
        )
    if use_uvloop == "auto":
        try:
            import uvloop  # noqa: F401 # pylint: disable=unused-variable

            use_uvloop = True
        except ImportError:
            use_uvloop = False

    external_addresses = pool_cls.get_external_addresses(
        address, n_process=n_process, ports=ports
    )
    actor_pool_config = ActorPoolConfig()
    actor_pool_config.add_metric_configs(kwargs.get("metrics", {}))
    # add main config
    process_index_gen = pool_cls.process_index_gen(address)
    main_process_index = next(process_index_gen)
    actor_pool_config.add_pool_conf(
        main_process_index,
        labels[0] if labels else None,
        pool_cls.gen_internal_address(main_process_index, external_addresses[0]),
        external_addresses[0],
        modules=modules,
        suspend_sigint=suspend_sigint,
        use_uvloop=use_uvloop,
        logging_conf=logging_conf,
        kwargs=kwargs,
    )
    # add sub configs
    for i in range(n_process):
        sub_process_index = next(process_index_gen)
        actor_pool_config.add_pool_conf(
            sub_process_index,
            labels[i + 1] if labels else None,
            pool_cls.gen_internal_address(sub_process_index, external_addresses[i + 1]),
            external_addresses[i + 1],
            env=envs[i] if envs else None,
            modules=modules,
            suspend_sigint=suspend_sigint,
            use_uvloop=use_uvloop,
            logging_conf=logging_conf,
            kwargs=kwargs,
        )

    pool: MainActorPoolType = await pool_cls.create(
        {
            "actor_pool_config": actor_pool_config,
            "process_index": main_process_index,
            "start_method": subprocess_start_method,
            "auto_recover": auto_recover,
            "on_process_down": on_process_down,
            "on_process_recover": on_process_recover,
        }
    )
    await pool.start()
    return pool



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .quantile import DataFrameQuantile


def _install():
    from ..core import SERIES_TYPE, DATAFRAME_TYPE
    from .quantile import quantile_series, quantile_dataframe
    from .corr import df_corr, df_corrwith, series_corr, series_autocorr

    for t in SERIES_TYPE:
        t.quantile = quantile_series
        t.corr = series_corr
        t.autocorr = series_autocorr

    for t in DATAFRAME_TYPE:
        t.quantile = quantile_dataframe
        t.corr = df_corr
        t.corrwith = df_corrwith


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict

try:
    from prometheus_client import Gauge as PGauge
except ImportError:
    PGauge = None

from ..metric import (
    AbstractMetric,
    AbstractCounter,
    AbstractGauge,
    AbstractMeter,
    AbstractHistogram,
)


class PrometheusMetricMixin(AbstractMetric):
    def _init(self):
        self._metric = (
            PGauge(self._name, self._description, self._tag_keys) if PGauge else None
        )

    def _record(self, value=1, tags: Optional[Dict[str, str]] = None):
        if self._metric:
            if tags:
                self._metric.labels(**tags).set(value)
            else:
                self._metric.set(value)


class Counter(PrometheusMetricMixin, AbstractCounter):
    pass


class Gauge(PrometheusMetricMixin, AbstractGauge):
    pass


class Meter(PrometheusMetricMixin, AbstractMeter):
    pass


class Histogram(PrometheusMetricMixin, AbstractHistogram):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
from typing import Dict, Union

from ..errors import ServerClosed
from .communication import Client
from .message import _MessageBase, ResultMessage, ErrorMessage, DeserializeMessageFailed
from .router import Router
from ...oscar.profiling import ProfilingData
from ...utils import Timer


result_message_type = Union[ResultMessage, ErrorMessage]


class ActorCaller:
    __slots__ = "_client_to_message_futures", "_clients"

    def __init__(self):
        self._client_to_message_futures: Dict[
            Client, Dict[bytes, asyncio.Future]
        ] = dict()
        self._clients: Dict[Client, asyncio.Task] = dict()

    async def get_client(self, router: Router, dest_address: str) -> Client:
        client = await router.get_client(dest_address, from_who=self)
        if client not in self._clients:
            self._clients[client] = asyncio.create_task(self._listen(client))
            self._client_to_message_futures[client] = dict()
        return client

    async def _listen(self, client: Client):
        while not client.closed:
            try:
                try:
                    message: _MessageBase = await client.recv()
                except (EOFError, ConnectionError, BrokenPipeError):
                    # remote server closed, close client and raise ServerClosed
                    try:
                        await client.close()
                    except (ConnectionError, BrokenPipeError):
                        # close failed, ignore it
                        pass
                    raise ServerClosed(
                        f"Remote server {client.dest_address} closed"
                    ) from None
                future = self._client_to_message_futures[client].pop(message.message_id)
                future.set_result(message)
            except DeserializeMessageFailed as e:  # pragma: no cover
                message_id = e.message_id
                future = self._client_to_message_futures[client].pop(message_id)
                future.set_exception(e)
            except Exception as e:  # noqa: E722  # pylint: disable=bare-except
                message_futures = self._client_to_message_futures.get(client)
                self._client_to_message_futures[client] = dict()
                for future in message_futures.values():
                    future.set_exception(copy.copy(e))
            finally:
                await asyncio.sleep(0)

        message_futures = self._client_to_message_futures.get(client)
        self._client_to_message_futures[client] = dict()
        error = ServerClosed(f"Remote server {client.dest_address} closed")
        for future in message_futures.values():
            future.set_exception(copy.copy(error))

    async def call(
        self,
        router: Router,
        dest_address: str,
        message: _MessageBase,
        wait: bool = True,
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        client = await self.get_client(router, dest_address)
        loop = asyncio.get_running_loop()
        wait_response = loop.create_future()
        self._client_to_message_futures[client][message.message_id] = wait_response

        with Timer() as timer:
            try:
                await client.send(message)
            except ConnectionError:
                try:
                    await client.close()
                except ConnectionError:
                    # close failed, ignore it
                    pass
                raise ServerClosed(f"Remote server {client.dest_address} closed")

            if not wait:
                r = wait_response
            else:
                r = await wait_response

        ProfilingData.collect_actor_call(message, timer.duration)
        return r

    async def stop(self):
        try:
            await asyncio.gather(*[client.close() for client in self._clients])
        except (ConnectionError, ServerClosed):
            pass
        self.cancel_tasks()

    def cancel_tasks(self):
        # cancel listening for all clients
        _ = [task.cancel() for task in self._clients.values()]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from functools import partial

from .. import opcodes
from ..core import ENTITY_TYPE, TILEABLE_TYPE, ChunkData
from ..core.custom_log import redirect_custom_log
from ..core.operand import ObjectOperand
from ..dataframe.core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE
from ..serialization.serializables import (
    FunctionField,
    ListField,
    DictField,
    BoolField,
    Int32Field,
)
from ..tensor.core import TENSOR_TYPE
from ..utils import (
    build_fetch_tileable,
    enter_current_session,
    find_objects,
    replace_objects,
    get_params_fields,
)
from .operands import RemoteOperandMixin


class RemoteFunction(RemoteOperandMixin, ObjectOperand):
    _op_type_ = opcodes.REMOTE_FUNCATION
    _op_module_ = "remote"

    _function = FunctionField("function")
    _function_args = ListField("function_args")
    _function_kwargs = DictField("function_kwargs")
    _retry_when_fail = BoolField("retry_when_fail")
    _n_output = Int32Field("n_output")

    def __init__(
        self,
        function=None,
        function_args=None,
        function_kwargs=None,
        retry_when_fail=None,
        n_output=None,
        **kw,
    ):
        super().__init__(
            _function=function,
            _function_args=function_args,
            _function_kwargs=function_kwargs,
            _retry_when_fail=retry_when_fail,
            _n_output=n_output,
            **kw,
        )

    @property
    def function(self):
        return self._function

    @property
    def function_args(self):
        return self._function_args

    @property
    def function_kwargs(self):
        return self._function_kwargs

    @property
    def retry_when_fail(self):
        return self._retry_when_fail

    @property
    def n_output(self):
        return self._n_output

    @property
    def output_limit(self):
        return self._n_output or 1

    @property
    def retryable(self) -> bool:
        return self._retry_when_fail

    @classmethod
    def _no_prepare(cls, tileable):
        return isinstance(
            tileable, (TENSOR_TYPE, DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE)
        )

    def _set_inputs(self, inputs):
        raw_inputs = getattr(self, "_inputs", None)
        super()._set_inputs(inputs)

        function_inputs = iter(inp for inp in self._inputs)
        mapping = {inp: new_inp for inp, new_inp in zip(inputs, self._inputs)}
        if raw_inputs is not None:
            for raw_inp in raw_inputs:
                if self._no_prepare(raw_inp):
                    if not isinstance(self._inputs[0], ChunkData):
                        # not in tile, set_inputs from tileable
                        mapping[raw_inp] = next(function_inputs)
                    else:
                        # in tile, set_inputs from chunk
                        mapping[raw_inp] = build_fetch_tileable(raw_inp)
                else:
                    mapping[raw_inp] = next(function_inputs)
        self._function_args = replace_objects(self._function_args, mapping)
        self._function_kwargs = replace_objects(self._function_kwargs, mapping)

    def __call__(self):
        find_inputs = partial(find_objects, types=ENTITY_TYPE)
        inputs = find_inputs(self._function_args) + find_inputs(self._function_kwargs)
        if self.n_output is None:
            return self.new_tileable(inputs)
        else:
            return self.new_tileables(
                inputs, kws=[dict(i=i) for i in range(self.n_output)]
            )

    @classmethod
    def tile(cls, op):
        outs = op.outputs
        chunk_op = op.copy().reset_key()

        chunk_inputs = []
        pure_depends = []
        executed = False
        for inp in op.inputs:
            if cls._no_prepare(inp):  # pragma: no cover
                if not executed:
                    # trigger execution
                    yield
                else:
                    executed = True
                # if input is tensor, DataFrame etc,
                # do not prepare data, because the data may be to huge,
                # and users can choose to fetch slice of the data themselves
                pure_depends.extend([True] * len(inp.chunks))
            else:
                pure_depends.extend([False] * len(inp.chunks))
            chunk_inputs.extend(inp.chunks)
        chunk_op._pure_depends = pure_depends
        # record tileable op key for chunk op
        chunk_op.tileable_op_key = op.key

        out_chunks = [list() for _ in range(len(outs))]
        chunk_kws = []
        for i, out in enumerate(outs):
            chunk_params = out.params
            chunk_params["index"] = ()
            chunk_params["i"] = i
            chunk_kws.append(chunk_params)
        chunks = chunk_op.new_chunks(chunk_inputs, kws=chunk_kws)
        for i, c in enumerate(chunks):
            out_chunks[i].append(c)

        kws = []
        for i, out in enumerate(outs):
            params = out.params
            params["chunks"] = out_chunks[i]
            params["nsplits"] = ()
            kws.append(params)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "RemoteFunction"):
        mapping = {
            inp: ctx[inp.key]
            for inp, is_pure_dep in zip(op.inputs, op.pure_depends)
            if not is_pure_dep
        }
        for to_search in [op.function_args, op.function_kwargs]:
            tileables = find_objects(to_search, TILEABLE_TYPE)
            for tileable in tileables:
                chunks = tileable.chunks
                fields = get_params_fields(chunks[0])
                metas = ctx.get_chunks_meta(
                    [chunk.key for chunk in chunks], fields=fields
                )
                for chunk, meta in zip(chunks, metas):
                    chunk.params = {field: meta[field] for field in fields}
                tileable.refresh_params()

        function = op.function
        function_args = replace_objects(op.function_args, mapping)
        function_kwargs = replace_objects(op.function_kwargs, mapping)

        result = function(*function_args, **function_kwargs)

        if op.n_output is None:
            ctx[op.outputs[0].key] = result
        else:
            if not isinstance(result, Iterable):
                raise TypeError(
                    f"Specifying n_output={op.n_output}, "
                    f"but result is not iterable, got {result}"
                )
            result = list(result)
            if len(result) != op.n_output:
                raise ValueError(
                    f"Length of return value should be {op.n_output}, "
                    f"got {len(result)}"
                )
            for out, r in zip(op.outputs, result):
                ctx[out.key] = r


def spawn(func, args=(), kwargs=None, retry_when_fail=False, n_output=None):
    """
    Spawn a function and return a Mars Object which can be executed later.

    Parameters
    ----------
    func : function
        Function to spawn.
    args: tuple
       Args to pass to function
    kwargs: dict
       Kwargs to pass to function
    retry_when_fail: bool, default False
       If True, retry when function failed.
    n_output: int
       Count of outputs for the function

    Returns
    -------
    Object
        Mars Object.

    Examples
    --------
    >>> import mars.remote as mr
    >>> def inc(x):
    >>>     return x + 1
    >>>
    >>> result = mr.spawn(inc, args=(0,))
    >>> result
    Object <op=RemoteFunction, key=e0b31261d70dd9b1e00da469666d72d9>
    >>> result.execute().fetch()
    1

    List of spawned functions can be converted to :class:`mars.remote.ExecutableTuple`,
    and `.execute()` can be called to run together.

    >>> results = [mr.spawn(inc, args=(i,)) for i in range(10)]
    >>> mr.ExecutableTuple(results).execute().fetch()
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Mars Object returned by :meth:`mars.remote.spawn` can be treated
    as arguments for other spawn functions.

    >>> results = [mr.spawn(inc, args=(i,)) for i in range(10)]   # list of spawned functions
    >>> def sum_all(xs):
            return sum(xs)
    >>> mr.spawn(sum_all, args=(results,)).execute().fetch()
    55

    inside a spawned function, new functions can be spawned.

    >>> def driver():
    >>>     results = [mr.spawn(inc, args=(i,)) for i in range(10)]
    >>>     return mr.ExecutableTuple(results).execute().fetch()
    >>>
    >>> mr.spawn(driver).execute().fetch()
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Mars tensor, DataFrame and so forth is available in spawned functions as well.

    >>> import mars.tensor as mt
    >>> def driver2():
    >>>     t = mt.random.rand(10, 10)
    >>>     return t.sum().to_numpy()
    >>>
    >>> mr.spawn(driver2).execute().fetch()
    52.47844223908132

    Argument of `n_output` can indicate that the spawned function will return multiple outputs.
    This is important when some of the outputs may be passed to different functions.

    >>> def triage(alist):
    >>>     ret = [], []
    >>>     for i in alist:
    >>>         if i < 0.5:
    >>>             ret[0].append(i)
    >>>         else:
    >>>             ret[1].append(i)
    >>>     return ret
    >>>
    >>> def sum_all(xs):
    >>>     return sum(xs)
    >>>
    >>> l = [0.4, 0.7, 0.2, 0.8]
    >>> la, lb = mr.spawn(triage, args=(l,), n_output=2)
    >>>
    >>> sa = mr.spawn(sum_all, args=(la,))
    >>> sb = mr.spawn(sum_all, args=(lb,))
    >>> mr.ExecutableTuple([sa, sb]).execute().fetch()
    >>> [0.6000000000000001, 1.5]
    """
    if not isinstance(args, tuple):
        args = [args]
    else:
        args = list(args)
    if kwargs is None:
        kwargs = dict()
    if not isinstance(kwargs, dict):
        raise TypeError("kwargs has to be a dict")

    op = RemoteFunction(
        function=func,
        function_args=args,
        function_kwargs=kwargs,
        retry_when_fail=retry_when_fail,
        n_output=n_output,
    )
    return op()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import List, Dict, Set

from ....typing import BandType
from ...core import NodeRole
from ..core import NodeStatus


class AbstractClusterAPI:
    @staticmethod
    def _calc_statuses(
        statuses: Set[NodeStatus] = None, exclude_statuses: Set[NodeStatus] = None
    ) -> Set[NodeStatus]:
        if statuses:
            return statuses
        elif exclude_statuses is not None:
            return set(NodeStatus.__members__.values()).difference(exclude_statuses)
        else:
            return {NodeStatus.READY}

    @abstractmethod
    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        """
        Get supervisor addresses

        Returns
        -------
        out
            list of supervisors
        """

    @abstractmethod
    async def watch_supervisors(self):
        """
        Watch supervisor addresses

        Returns
        -------
        out
            generator of list of supervisors
        """

    @abstractmethod
    async def watch_nodes(
        self,
        role: NodeRole,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> List[Dict[str, Dict]]:
        """
        Watch changes of workers

        Returns
        -------
        out: List[Dict[str, Dict]]
            dict of worker resources by addresses and bands
        """

    @abstractmethod
    async def get_nodes_info(
        self,
        nodes: List[str] = None,
        role: NodeRole = None,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        """
        Get worker info

        Parameters
        ----------
        nodes
            address of nodes
        role
            roles of nodes
        env
            receive env info
        resource
            receive resource info
        detail
            receive detail info

        Returns
        -------
        out: Dict
            info of worker
        """

    @abstractmethod
    async def get_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> Dict[BandType, int]:
        """
        Get all bands that can be used for computation.

        Returns
        -------
        band_to_slots : dict
            Band to n_slot.
        """

    @abstractmethod
    async def watch_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        """
        Watch all bands that can be used for computation.

        Returns
        -------
        band_to_slots : dict
            Band to n_slot.
        """

    @abstractmethod
    async def get_mars_versions(self) -> List[str]:
        """
        Get versions used in current Mars cluster

        Returns
        -------
        version_list : list
            List of versions
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict

from ....utils import lazy_import
from ..metric import (
    AbstractMetric,
    AbstractCounter,
    AbstractGauge,
    AbstractMeter,
    AbstractHistogram,
)

ray_metrics = lazy_import("ray.util.metrics")

# Note: Gauge `record` method is deprecated in ray 1.3.0 version, so here
# make it compatible with the old and new ray versions.
RAY_GAUGE_SET_AVAILABLE = (
    True if ray_metrics and hasattr(ray_metrics.Gauge, "set") else False
)


class RayMetricMixin(AbstractMetric):
    def _init(self):
        if ray_metrics is not None:  # pragma: no branch
            self._metric = ray_metrics.Gauge(
                self._name, self._description, self._tag_keys
            )

    def _record(self, value=1, tags: Optional[Dict[str, str]] = None):
        if RAY_GAUGE_SET_AVAILABLE:
            self._metric.set(value, tags)
        elif ray_metrics is not None:  # pragma: no branch
            self._metric.record(value, tags)


class Counter(RayMetricMixin, AbstractCounter):
    pass


class Gauge(RayMetricMixin, AbstractGauge):
    pass


class Meter(RayMetricMixin, AbstractMeter):
    pass


class Histogram(RayMetricMixin, AbstractHistogram):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="unary")
class TensorCosh(TensorUnaryOp):
    _op_type_ = OperandDef.COSH
    _func_name = "cosh"


@infer_dtype(np.cosh)
def cosh(x, out=None, where=None, **kwargs):
    """
    Hyperbolic cosine, element-wise.

    Equivalent to ``1/2 * (mt.exp(x) + mt.exp(-x))`` and ``mt.cos(1j*x)``.

    Parameters
    ----------
    x : array_like
        Input tensor.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : Tensor
        Output array of same shape as `x`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.cosh(0).execute()
    1.0

    The hyperbolic cosine describes the shape of a hanging cable:

    >>> import matplotlib.pyplot as plt
    >>> x = mt.linspace(-4, 4, 1000)
    >>> plt.plot(x.execute(), mt.cosh(x).execute())
    >>> plt.show()
    """
    op = TensorCosh(**kwargs)
    return op(x, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import numpy as np

try:
    import scipy.sparse as sps
except ImportError:  # pragma: no cover
    sps = None

from .core import Serializer, buffered, serialize, deserialize


class CsrMatrixSerializer(Serializer):
    serializer_name = "sps.csr_matrix"

    @buffered
    def serialize(self, obj: Any, context: Dict):
        data_header, data_buffers = serialize(obj.data)
        idx_header, idx_buffers = serialize(obj.indices)
        indptr_header, indptr_buffers = serialize(obj.indptr)
        header = {
            "data_header": data_header,
            "data_buf_num": len(data_buffers),
            "idx_header": idx_header,
            "idx_buf_num": len(idx_buffers),
            "indptr_header": indptr_header,
            "shape": list(obj.shape),
        }
        return header, data_buffers + idx_buffers + indptr_buffers

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        data_buf_num = header["data_buf_num"]
        idx_buf_num = header["idx_buf_num"]
        data_buffers = buffers[:data_buf_num]
        idx_buffers = buffers[data_buf_num : data_buf_num + idx_buf_num]
        indptr_buffers = buffers[data_buf_num + idx_buf_num :]

        data = deserialize(header["data_header"], data_buffers)
        indices = deserialize(header["idx_header"], idx_buffers)
        indptr = deserialize(header["indptr_header"], indptr_buffers)
        shape = tuple(header["shape"])

        empty_arr = np.zeros(0, dtype=data.dtype)

        target_csr = sps.coo_matrix(
            (empty_arr, (empty_arr,) * 2), dtype=data.dtype, shape=shape
        ).tocsr()
        target_csr.data, target_csr.indices, target_csr.indptr = data, indices, indptr
        return target_csr


if sps:  # pragma: no branch
    CsrMatrixSerializer.register(sps.csr_matrix)
